import csv
from copy import copy
from dataclasses import dataclass
from functools import partial, wraps
from pathlib import Path
from statistics import mean
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
    cast,
)

import numpy as np
import pandas as pd
import torch
from aislib.misc_utils import ensure_path_exists
from lifelines.utils import concordance_index
from scipy.special import softmax
from scipy.stats import pearsonr
from sklearn.metrics import (
    average_precision_score,
    cohen_kappa_score,
    explained_variance_score,
    f1_score,
    mean_absolute_error,
    precision_score,
    r2_score,
    recall_score,
    roc_auc_score,
)
from sklearn.preprocessing import KBinsDiscretizer, StandardScaler, label_binarize
from torch.linalg import vector_norm

from eir.data_load.data_utils import get_output_info_generator
from eir.setup.schemas import (
    OutputConfig,
    SurvivalOutputTypeConfig,
    TabularOutputTypeConfig,
)
from eir.target_setup.target_label_setup import MissingTargetsInfo
from eir.target_setup.target_setup_utils import IdentityTransformer
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.label_setup import (  # noqa: F401
        al_label_transformers,
        al_label_transformers_object,
        al_target_columns,
    )
    from eir.models.input.omics.omics_models import al_omics_models  # noqa: F401
    from eir.models.meta.meta_utils import FeatureExtractorProtocolWithL1
    from eir.setup.output_setup import al_output_objects_as_dict
    from eir.train import Experiment, al_criteria_dict  # noqa: F401
    from eir.train_utils.step_logic import al_training_labels_target
    from eir.train_utils.train_handlers import HandlerConfig

# aliases
# output_name -> target_name -> metric name: value
al_step_metric_dict = dict[str, dict[str, dict[str, float]]]

logger = get_logger(name=__name__)


class MetricFunctionProtocol(Protocol):
    def __call__(
        self,
        outputs: np.ndarray,
        labels: np.ndarray,
        column_name: str,
        output_name: str,
    ) -> float: ...


class SurvivalMetricFunctionProtocol(Protocol):
    def __call__(
        self,
        outputs: np.ndarray,
        labels: np.ndarray,
        times: np.ndarray,
        event_name: str,
        time_name: str,
        output_name: str,
    ) -> float: ...


class AverageMetricFunctionProtocol(Protocol):
    def __call__(
        self,
        metric_dict: al_step_metric_dict,
        output_name: str,
        column_name: str,
    ) -> float: ...


al_averaging_functions_dict = Dict[str, AverageMetricFunctionProtocol]
al_metric_record_dict = Dict[
    str,
    Tuple["MetricRecord", ...] | "al_averaging_functions_dict" | "GeneralMetricInfo",
]

al_cat_metric_choices = Sequence[
    Literal[
        "mcc",
        "acc",
        "roc-auc-macro",
        "ap-macro",
        "f1-macro",
        "precision-macro",
        "recall-macro",
        "cohen-kappa",
    ]
]
al_con_metric_choices = Sequence[
    Literal[
        "r2",
        "pcc",
        "loss",
        "rmse",
        "mae",
        "mape",
        "explained-variance",
    ]
]

all_survival_metric_choices = Sequence[Literal["c-index"]]


@dataclass()
class MetricRecord:
    name: str
    function: MetricFunctionProtocol | SurvivalMetricFunctionProtocol
    output_type: str = "supervised"
    only_val: bool = True
    minimize_goal: bool = False


def calculate_batch_metrics(
    outputs_as_dict: "al_output_objects_as_dict",
    outputs: Dict[str, Dict[str, torch.Tensor]],
    labels: "al_training_labels_target",
    mode: str,
    metric_record_dict: al_metric_record_dict,
) -> al_step_metric_dict:
    assert mode in ["val", "train"]

    general_info = metric_record_dict["general_metric_info"]
    assert isinstance(general_info, GeneralMetricInfo)
    if mode == "train" and general_info.all_are_val_only:
        return general_info.base_metric_structure

    target_columns_gen = get_output_info_generator(outputs_as_dict=outputs_as_dict)

    master_metric_dict: al_step_metric_dict = {}

    for output_name, output_target_type, target_name in target_columns_gen:
        cur_metric_dict: dict[str, float] = {}

        if output_name not in master_metric_dict:
            master_metric_dict[output_name] = {}

        if output_target_type == "general":
            master_metric_dict[output_name][target_name] = cur_metric_dict

        elif output_target_type in ("con", "cat"):
            cur_output_object = outputs_as_dict[output_name]
            cur_output_type = cur_output_object.output_config.output_info.output_type
            assert cur_output_type == "tabular"

            al_record = Tuple[MetricRecord, ...]
            cur_records = metric_record_dict[output_target_type]
            assert isinstance(cur_records, tuple)
            cur_metric_records: al_record = cur_records

            cur_outputs = outputs[output_name][target_name]
            cur_target_labels = labels[output_name][target_name]

            filtered = filter_tabular_missing_targets(
                outputs=cur_outputs,
                target_labels=cur_target_labels,
                ids=[],
                target_type=output_target_type,
            )

            cur_outputs_np = general_torch_to_numpy(tensor=filtered.model_outputs)
            cur_labels_np = general_torch_to_numpy(tensor=filtered.target_labels)
            if output_target_type == "cat":
                cur_labels_np = cur_labels_np.astype(int)

            for metric_record in cur_metric_records:

                if metric_record.output_type != "supervised":
                    continue

                if metric_record.only_val and mode == "train":
                    continue

                cur_key = f"{output_name}_{target_name}_{metric_record.name}"
                cur_fun = cast(MetricFunctionProtocol, metric_record.function)
                cur_value = cur_fun(
                    outputs=cur_outputs_np,
                    labels=cur_labels_np,
                    column_name=target_name,
                    output_name=output_name,
                )

                cur_metric_dict[cur_key] = cur_value

            master_metric_dict[output_name][target_name] = cur_metric_dict

        elif output_target_type == "survival":
            cur_output_object = outputs_as_dict[output_name]
            cur_output_type = cur_output_object.output_config.output_info.output_type
            assert cur_output_type == "survival"

            cur_records = metric_record_dict[output_target_type]
            assert isinstance(cur_records, tuple)
            cur_metric_records = cur_records

            cur_outputs = outputs[output_name][target_name]
            cur_events = labels[output_name][target_name]

            cur_oti = cur_output_object.output_config.output_type_info
            assert isinstance(cur_oti, SurvivalOutputTypeConfig)

            cur_time_name = cur_oti.time_column
            cur_times = labels[output_name][cur_time_name]

            filtered = filter_survival_missing_targets(
                model_outputs=cur_outputs,
                events=cur_events,
                times=cur_times,
                cur_ids=[],
            )

            cur_outputs_np = general_torch_to_numpy(tensor=filtered.model_outputs)
            cur_labels_np = general_torch_to_numpy(tensor=filtered.events)
            cur_labels_np = cur_labels_np.astype(int)
            cur_times_np = general_torch_to_numpy(tensor=filtered.times)

            for metric_record in cur_metric_records:

                if metric_record.output_type != "survival":
                    continue

                if metric_record.only_val and mode == "train":
                    continue

                cur_key = f"{output_name}_{target_name}_{metric_record.name}"
                cur_fun = cast(SurvivalMetricFunctionProtocol, metric_record.function)
                cur_value = cur_fun(
                    outputs=cur_outputs_np,
                    labels=cur_labels_np,
                    times=cur_times_np,
                    event_name=target_name,
                    output_name=output_name,
                    time_name=cur_time_name,
                )

                cur_metric_dict[cur_key] = cur_value

            master_metric_dict[output_name][target_name] = cur_metric_dict
        else:
            raise NotImplementedError()

    return master_metric_dict


def add_loss_to_metrics(
    outputs_as_dict: "al_output_objects_as_dict",
    losses: Dict[str, Dict[str, torch.Tensor]],
    metric_dict: al_step_metric_dict,
) -> al_step_metric_dict:
    target_columns_gen = get_output_info_generator(outputs_as_dict=outputs_as_dict)
    metric_dict_copy = copy(metric_dict)

    for output_name, column_type, column_name in target_columns_gen:
        cur_metric_dict = metric_dict_copy[output_name][column_name]
        cur_key = f"{output_name}_{column_name}_loss"
        cur_metric_dict[cur_key] = losses[output_name][column_name].item()

    return metric_dict_copy


def add_multi_task_average_metrics(
    batch_metrics_dict: al_step_metric_dict,
    outputs_as_dict: "al_output_objects_as_dict",
    loss: float,
    performance_average_functions: Optional["al_averaging_functions_dict"],
) -> al_step_metric_dict:
    average = {
        "average": {
            "loss-average": loss,
        }
    }

    if performance_average_functions is not None:
        average_performance = average_performances_across_tasks(
            metric_dict=batch_metrics_dict,
            outputs_as_dict=outputs_as_dict,
            performance_calculation_functions=performance_average_functions,
        )
        average["average"]["perf-average"] = average_performance

    batch_metrics_dict["average"] = average

    return batch_metrics_dict


def average_performances_across_tasks(
    metric_dict: al_step_metric_dict,
    outputs_as_dict: "al_output_objects_as_dict",
    performance_calculation_functions: "al_averaging_functions_dict",
) -> float:
    target_columns_gen = get_output_info_generator(
        outputs_as_dict=outputs_as_dict,
    )

    all_metrics = []

    for output_name, column_type, column_name in target_columns_gen:
        cur_metric_func = performance_calculation_functions[column_type]

        cur_value = cur_metric_func(
            metric_dict=metric_dict, output_name=output_name, column_name=column_name
        )

        if not np.isnan(cur_value):
            all_metrics.append(cur_value)

    average = np.array(all_metrics).mean()

    return average


def handle_empty(default_value: float, metric_name: Optional[str] = None):
    """
    This can happen when modelling on multiple outputs, which can vary in their
    sparsity, and by chance some outputs are empty in a batch.
    """

    def decorator(func):
        logged = False

        @wraps(func)
        def wrapper(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs):
            nonlocal logged
            if (outputs.size == 0 or labels.size == 0) and not logged:
                logger.info(
                    f"Empty inputs encountered in "
                    f"{metric_name or func.__name__}, "
                    f"returning default value: {default_value}"
                )
                logged = True
            if outputs.size == 0 or labels.size == 0:
                return default_value
            return func(outputs, labels, *args, **kwargs)

        return wrapper

    return decorator


def handle_class_mismatch(default_value: float, metric_name: Optional[str] = None):
    """
    Decorator to handle cases where the number of unique classes in 'labels'
    does not match the number of columns in 'outputs'. This scenario can occur
    in multi-class classification tasks with sparse or imbalanced outputs.
    """

    def decorator(func):
        logged = False

        @wraps(func)
        def wrapper(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs):
            nonlocal logged
            unique_classes = np.unique(labels)
            # For binary classification as CE we have this special case
            if outputs.shape[1] == 2 and len(unique_classes) == 1:
                pass
            # For binary classification as BCE we have this special case
            elif outputs.shape[1] == 1:
                pass
            elif len(unique_classes) != outputs.shape[1]:
                if not logged:
                    logger.info(
                        f"Class mismatch encountered in "
                        f"{metric_name or func.__name__}, "
                        f"expected number of classes: {outputs.shape[1]}, "
                        f"found unique classes in labels: {len(unique_classes)}. "
                        f"Returning default value: {default_value}"
                    )
                    logged = True
                return default_value
            return func(outputs, labels, *args, **kwargs)

        return wrapper

    return decorator


@handle_empty(default_value=np.nan, metric_name="MCC")
def calc_mcc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    prediction = parse_cat_pred(outputs=outputs)

    labels = labels.astype(int)

    num_classes = int(max(np.max(labels), np.max(prediction))) + 1

    conf_matrix = np.zeros((num_classes, num_classes), dtype=np.int64)
    np.add.at(conf_matrix, (labels, prediction), 1)

    t_sum = conf_matrix.sum(axis=1, dtype=np.float64)
    p_sum = conf_matrix.sum(axis=0, dtype=np.float64)
    n_correct = np.sum(np.diag(conf_matrix), dtype=np.float64)
    n_samples = np.sum(conf_matrix, dtype=np.float64)

    cov_ytyp = n_correct * n_samples - np.dot(t_sum, p_sum)
    cov_ypyp = n_samples**2 - np.dot(p_sum, p_sum)
    cov_ytyt = n_samples**2 - np.dot(t_sum, t_sum)

    if cov_ypyp * cov_ytyt == 0:
        return 0.0
    else:
        return cov_ytyp / np.sqrt(cov_ytyt * cov_ypyp)


@handle_class_mismatch(default_value=np.nan, metric_name="ROC-AUC")
@handle_empty(default_value=np.nan, metric_name="ROC-AUC")
def calc_roc_auc_ovo(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    """
    TODO:   In rare scenarios, we might run into the issue of not having all labels
            represented in the labels array (i.e. labels were in train, but not in
            valid). This is not a problem for metrics like MCC / accuracy, but we
            will have to account for this here and in the AP calculation, possibly
            by ignoring columns in outputs and label_binarize outputs where the columns
            returned from label_binarize are all 0.
    """

    average = "macro"

    if outputs.shape[1] > 2:
        outputs = softmax(x=outputs, axis=1)
    elif outputs.shape[1] == 2:
        outputs = outputs[:, 1]

    roc_auc = roc_auc_score(
        y_true=labels,
        y_score=outputs,
        average=average,
        multi_class="ovo",
    )
    return roc_auc


@handle_class_mismatch(default_value=np.nan, metric_name="AP")
@handle_empty(default_value=np.nan, metric_name="AP")
def calc_average_precision(
    outputs: np.ndarray, labels: np.ndarray, *args, **kwargs
) -> float:
    average = "macro"

    if outputs.shape[1] > 2:
        labels = label_binarize(y=labels, classes=sorted(np.unique(labels)))
    elif outputs.shape[1] == 2:
        outputs = outputs[:, 1]

    average_precision = average_precision_score(
        y_true=labels,
        y_score=outputs,
        average=average,
    )

    return average_precision


def parse_cat_pred(outputs: np.ndarray) -> np.ndarray:
    # Single logit case
    if outputs.shape[1] == 1:
        pred = (outputs > 0).astype(int).squeeze()
    # Multi-class case
    else:
        pred = np.argmax(outputs, axis=1)

    return pred


@handle_empty(default_value=np.nan, metric_name="ACC")
def calc_acc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    pred = parse_cat_pred(outputs=outputs)
    accuracy = np.mean(pred == labels)

    return accuracy


@handle_empty(default_value=np.nan, metric_name="F1-MACRO")
def calc_f1_score_macro(
    outputs: np.ndarray, labels: np.ndarray, *args, **kwargs
) -> float:
    pred = parse_cat_pred(outputs=outputs)
    f1 = f1_score(y_true=labels, y_pred=pred, average="macro")
    return f1


@handle_empty(default_value=np.nan, metric_name="PRECISION-MACRO")
def calc_precision_macro(
    outputs: np.ndarray, labels: np.ndarray, *args, **kwargs
) -> float:
    pred = parse_cat_pred(outputs=outputs)
    precision = precision_score(y_true=labels, y_pred=pred, average="macro")
    return precision


@handle_empty(default_value=np.nan, metric_name="RECALL-MACRO")
def calc_recall_macro(
    outputs: np.ndarray, labels: np.ndarray, *args, **kwargs
) -> float:
    pred = parse_cat_pred(outputs=outputs)
    recall = recall_score(y_true=labels, y_pred=pred, average="macro")
    return recall


@handle_empty(default_value=np.nan, metric_name="COHEN-KAPPA")
def calc_cohen_kappa(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    pred = parse_cat_pred(outputs=outputs)
    kappa = cohen_kappa_score(y1=labels, y2=pred)
    return kappa


@handle_empty(default_value=np.nan, metric_name="PCC")
def calc_pcc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    if len(outputs) < 2:
        return 0.0

    pcc = pearsonr(x=labels.squeeze(), y=outputs.squeeze())[0]
    return pcc


@handle_empty(default_value=np.nan, metric_name="R2")
def calc_r2(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    if len(outputs) < 2:
        return 0.0

    r2 = r2_score(y_true=labels.squeeze(), y_pred=outputs.squeeze())
    return r2


@handle_empty(default_value=np.nan, metric_name="RMSE")
def calc_rmse(
    outputs: np.ndarray,
    labels: np.ndarray,
    target_transformers: Dict[str, Dict[str, StandardScaler]],
    output_name: str,
    column_name: str,
    *args,
    **kwargs,
) -> float:
    cur_target_transformer = target_transformers[output_name][column_name]

    labels_2d = labels.reshape(-1, 1)
    outputs_2d = outputs.reshape(-1, 1)

    mean_ = cur_target_transformer.mean_
    scale_ = cur_target_transformer.scale_

    labels = (labels_2d * scale_ + mean_).squeeze()
    predictions = (outputs_2d * scale_ + mean_).squeeze()

    if np.shape(labels):
        rmse = np.sqrt(np.mean((labels - predictions) ** 2))
    else:
        rmse = np.sqrt((labels - predictions) ** 2)

    return rmse


@handle_empty(default_value=np.nan, metric_name="MAE")
def calc_mae(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    mae = mean_absolute_error(y_true=labels.squeeze(), y_pred=outputs.squeeze())
    return mae


@handle_empty(default_value=np.nan, metric_name="MAPE")
def calc_mape(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    labels, outputs = labels.squeeze(), outputs.squeeze()
    mask = labels != 0
    return np.mean(np.abs((labels[mask] - outputs[mask]) / labels[mask])) * 100


@handle_empty(default_value=np.nan, metric_name="EXPLAINED-VARIANCE")
def calc_explained_variance(
    outputs: np.ndarray, labels: np.ndarray, *args, **kwargs
) -> float:
    ev = explained_variance_score(y_true=labels.squeeze(), y_pred=outputs.squeeze())
    return ev


@handle_empty(default_value=np.nan, metric_name="C-INDEX")
def calc_survival_c_index(
    outputs: np.ndarray,
    labels: np.ndarray,
    times: np.ndarray,
    target_transformers: dict[str, dict[str, KBinsDiscretizer | IdentityTransformer]],
    output_name: str,
    event_name: str,
    time_name: str,
    *args,
    **kwargs,
) -> float:
    """
    Note that for CoxPH models we just have the output being a single value quantifying
    risk predicted by the model for each sample. Whereas for discrete survival models
    we take the last bin (i.e., last time point) as the overall risk predicted by the
    model.
    """

    def sigmoid(x):
        return 1 / (1 + np.exp(-x))

    target_transformer = target_transformers[output_name][time_name]
    model_type = "discrete"
    if isinstance(target_transformer, IdentityTransformer):
        model_type = "cox"

    events = labels

    times = target_transformer.inverse_transform(times.reshape(-1, 1))
    times = times.flatten()

    # # Convert model outputs (risk) to survival scores
    # higher survival score = longer predicted survival time
    if model_type == "discrete":
        hazards = sigmoid(outputs)
        survival_scores = -hazards[:, -1]
    else:
        survival_scores = -outputs

    c_index = concordance_index(
        event_times=times,
        predicted_scores=survival_scores,
        event_observed=events,
    )

    return c_index


class LogEmptyLossProtocol(Protocol):
    def __call__(self, output_name: str, output_head_name: str) -> None: ...


def log_empty_loss_once() -> LogEmptyLossProtocol:
    logged_combinations = set()

    def log(output_name: str, output_head_name: str) -> None:
        if (output_name, output_head_name) not in logged_combinations:
            logger.info(
                f"Empty output batch encountered for {output_name},"
                f" {output_head_name}; "
                f"setting loss to NaN for this batch and future empty batches. "
                f"Empty batches will not be used for training."
            )
            logged_combinations.add((output_name, output_head_name))

    return log


def calculate_prediction_losses(
    criteria: "al_criteria_dict",
    inputs: dict[str, dict[str, torch.Tensor]],
    targets: dict[str, dict[str, torch.Tensor]],
) -> dict[str, dict[str, torch.Tensor]]:
    """
    We check for empty output tensors and log a warning if we encounter them.
    This can happen when modelling on multiple outputs, which can vary in their
    sparsity, and by chance some outputs are empty in a batch.
    """
    losses_dict: dict[str, dict[str, torch.Tensor]] = {}

    for output_name, target_dict in targets.items():
        cur_inputs = inputs[output_name]
        cur_targets = targets[output_name]
        cur_criterion = criteria[output_name]
        losses_dict[output_name] = {}

        cur_losses = cur_criterion(cur_inputs, cur_targets)
        for name, loss in cur_losses.items():
            losses_dict[output_name][name] = loss

    return losses_dict


def aggregate_losses(losses_dict: dict[str, dict[str, torch.Tensor]]) -> torch.Tensor:
    losses_values = []
    for output_name, targets_for_output_dict in losses_dict.items():
        for loss in targets_for_output_dict.values():
            losses_values.append(loss)

    if not losses_values:
        return torch.tensor(np.nan, requires_grad=True)

    average_loss = torch.nanmean(torch.stack(losses_values))
    return average_loss


def get_uncertainty_loss_hook(
    output_configs: Sequence[OutputConfig],
    device: str,
) -> tuple[Callable, dict[str, "UncertaintyMultiTaskLoss"]]:
    uncertainty_loss_modules: dict[str, UncertaintyMultiTaskLoss] = {}
    for output_config in output_configs:
        if output_config.output_info.output_type != "tabular":
            continue

        output_type_info = output_config.output_type_info
        assert isinstance(output_type_info, TabularOutputTypeConfig)

        if not output_type_info.uncertainty_weighted_mt_loss:
            continue

        logger.info(
            f"Adding uncertainty loss for {output_config.output_info.output_name}."
        )

        cur_target_cat_columns = list(output_type_info.target_cat_columns)
        cur_target_con_columns = list(output_type_info.target_con_columns)
        loss_module = UncertaintyMultiTaskLoss(
            target_cat_columns=cur_target_cat_columns,
            target_con_columns=cur_target_con_columns,
            device=device,
        )
        uncertainty_loss_modules[output_config.output_info.output_name] = loss_module

    if len(uncertainty_loss_modules) == 0:
        raise ValueError("Expected at least one uncertainty loss module.")

    hook = partial(
        hook_add_uncertainty_loss, uncertainty_modules=uncertainty_loss_modules
    )

    return hook, uncertainty_loss_modules


def hook_add_uncertainty_loss(
    state: Dict,
    uncertainty_modules: Dict[str, "UncertaintyMultiTaskLoss"],
    loss_key: str = "per_target_train_losses",
    *args,
    **kwargs,
) -> Dict[str, Dict[str, torch.Tensor]]:
    """
    Note that we only update the relevant losses in the base dict.
    """

    base_losses_dict = state[loss_key]
    updated_losses = copy(base_losses_dict)

    for output_name, module in uncertainty_modules.items():
        cur_module = uncertainty_modules[output_name]
        cur_loss_dict = base_losses_dict[output_name]
        cur_uncertainty_losses = cur_module(losses_dict=cur_loss_dict)
        updated_losses[output_name] = cur_uncertainty_losses

    state_updates = {loss_key: updated_losses}

    return state_updates


from typing import Dict, List

import torch
import torch.nn as nn


class UncertaintyMultiTaskLoss(nn.Module):
    def __init__(
        self,
        target_cat_columns: List[str],
        target_con_columns: List[str],
        device: str,
    ):
        super().__init__()

        self.target_cat_columns = target_cat_columns
        self.target_con_columns = target_con_columns
        self.device = device

        self._construct_params(
            cur_target_columns=self.target_cat_columns + self.target_con_columns,
            device=self.device,
        )

    def _construct_params(self, cur_target_columns: List[str], device: str):
        for column_name in cur_target_columns:
            param = nn.Parameter(torch.zeros(1, device=device), requires_grad=True)
            self.register_parameter(f"log_var_{column_name}", param)

    def _calc_uncertainty_loss(
        self,
        name: str,
        loss_value: torch.Tensor,
    ) -> torch.Tensor:
        log_var = getattr(self, f"log_var_{name}")
        scalar = 2.0 if name in self.target_cat_columns else 1.0

        precision = torch.exp(-log_var)
        loss = scalar * torch.sum(precision * loss_value + log_var)

        return loss

    def forward(self, losses_dict: Dict) -> Dict[str, torch.Tensor]:
        losses_uncertain = {}

        for target_column, loss_value_base in losses_dict.items():
            loss_value_uncertain = self._calc_uncertainty_loss(
                name=target_column,
                loss_value=loss_value_base,
            )
            losses_uncertain[target_column] = loss_value_uncertain

        return losses_uncertain


def hook_add_l1_loss(
    experiment: "Experiment",
    state: Dict,
    loss_key: str = "loss",
    *args,
    **kwargs,
) -> Dict:
    """
    TODO: Do the validation outside of the actual hook.
    """
    model_configs = experiment.inputs

    l1_loss = torch.tensor(0.0, device=experiment.configs.gc.be.device)
    for input_name, input_module in experiment.model.input_modules.items():
        cur_model_config = model_configs[input_name].input_config.model_config
        cur_model_init_config = cur_model_config.model_init_config

        current_l1 = getattr(cur_model_init_config, "l1", None)
        has_l1_weights = hasattr(input_module, "l1_penalized_weights")

        if current_l1 and not has_l1_weights:
            raise AttributeError(
                f"Module {input_module} for input name {input_name} does not have"
                f"l1_penalized_weights attribute."
            )

        if has_l1_weights and current_l1:
            input_module_with_l1 = cast("FeatureExtractorProtocolWithL1", input_module)
            cur_l1_loss = get_model_l1_loss(
                model=input_module_with_l1, l1_weight=current_l1
            )
            l1_loss += cur_l1_loss

    updated_loss = state[loss_key] + l1_loss

    state_updates = {loss_key: updated_loss}

    return state_updates


def get_model_l1_loss(
    model: "FeatureExtractorProtocolWithL1", l1_weight: float
) -> torch.Tensor:
    l1_loss = calc_l1_loss(
        weight_tensor=model.l1_penalized_weights, l1_weight=l1_weight
    )
    return l1_loss


def calc_l1_loss(weight_tensor: torch.Tensor, l1_weight: float):
    l1_loss = vector_norm(weight_tensor, ord=1, dim=None) * l1_weight
    return l1_loss


def persist_metrics(
    handler_config: "HandlerConfig",
    metrics_dict: "al_step_metric_dict",
    iteration: int,
    write_header: bool,
    prefixes: Dict[str, str],
    writer_funcs: Union[None, Dict[str, Dict[str, Callable]]] = None,
) -> None:

    hc = handler_config
    exp = handler_config.experiment
    gc = exp.configs.global_config

    target_generator = get_output_info_generator(outputs_as_dict=exp.outputs)

    metrics_files = get_metrics_files(
        target_generator=target_generator,
        run_folder=hc.run_folder,
        train_or_val_target_prefix=f"{prefixes['metrics']}",
        detail_level=gc.evaluation_checkpoint.saved_result_detail_level,
    )

    if write_header:
        _ensure_metrics_paths_exists(metrics_files=metrics_files)

    for output_name, target_and_file_dict in metrics_files.items():
        for target_name, target_history_file in target_and_file_dict.items():
            cur_metric_dict = metrics_dict[output_name][target_name]

            if writer_funcs:
                cur_func = writer_funcs[output_name][target_name]
            else:
                cur_func = get_buffered_metrics_writer(buffer_interval=1)

            cur_func(
                filepath=target_history_file,
                metrics=cur_metric_dict,
                iteration=iteration,
                write_header=write_header,
            )


def get_metrics_files(
    target_generator: Generator[Tuple[str, str, str], None, None],
    run_folder: Path,
    train_or_val_target_prefix: str,
    detail_level: int,
) -> dict[str, dict[str, Path]]:
    assert train_or_val_target_prefix in ["validation_", "train_"]

    path_dict: dict[str, dict[str, Path]] = {}
    for output_name, column_type, target_column in target_generator:
        if detail_level <= 1:
            continue

        if output_name not in path_dict:
            path_dict[output_name] = {}

        cur_file_name = train_or_val_target_prefix + target_column + "_history.log"
        cur_path = Path(
            run_folder,
            "results",
            output_name,
            target_column,
            cur_file_name,
        )
        path_dict[output_name][target_column] = cur_path

    average_loss_training_metrics_file = get_average_history_filepath(
        run_folder=run_folder, train_or_val_target_prefix=train_or_val_target_prefix
    )
    path_dict["average"] = {"average": average_loss_training_metrics_file}

    return path_dict


def get_average_history_filepath(
    run_folder: Path, train_or_val_target_prefix: str
) -> Path:
    assert train_or_val_target_prefix in ["validation_", "train_"]
    metrics_file_path = run_folder / f"{train_or_val_target_prefix}average_history.log"
    return metrics_file_path


def _ensure_metrics_paths_exists(metrics_files: Dict[str, Dict[str, Path]]) -> None:
    for output_name, target_and_file_dict in metrics_files.items():
        for path in target_and_file_dict.values():
            ensure_path_exists(path=path)


def get_buffered_metrics_writer(buffer_interval: int):
    """
    One might be tempted to use "a" mode here to append metrics to the log
    file instead of read-modify-write. However, some cloud storage platforms
    (e.g., S3) do not support appending (i.e., modifying) to files, raising
    OSError 95: Operation not supported.
    """
    buffer: list[dict[str, float]] = []

    def append_metrics_to_file(
        filepath: Path,
        metrics: dict[str, float],
        iteration: int,
        write_header: bool = False,
    ) -> None:
        nonlocal buffer

        dict_to_write = {"iteration": iteration, **metrics}
        buffer.append(dict_to_write)

        if not (iteration % buffer_interval == 0 or write_header):
            return

        fieldnames = ["iteration"] + sorted(metrics.keys())

        existing_rows = []
        if filepath.exists():
            with open(filepath, "r") as f:
                reader = csv.DictReader(f)
                existing_rows = [row for row in reader]

        all_rows = existing_rows + buffer

        with open(filepath, "w", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(all_rows)

        buffer.clear()

    return append_metrics_to_file


def read_metrics_history_file(file_path: Path) -> pd.DataFrame:
    df = pd.read_csv(file_path, index_col="iteration")

    return df


def get_metrics_dataframes(
    results_dir: Path, target_string: str
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    train_history_path = read_metrics_history_file(
        results_dir / f"train_{target_string}_history.log"
    )
    valid_history_path = read_metrics_history_file(
        results_dir / f"validation_{target_string}_history.log"
    )

    return train_history_path, valid_history_path


def get_available_supervised_metrics() -> (
    Tuple[Dict[str, MetricRecord], Dict[str, MetricRecord]]
):
    cat_metrics: Dict[str, MetricRecord] = {
        "mcc": MetricRecord(
            name="mcc",
            function=calc_mcc,
        ),
        "acc": MetricRecord(
            name="acc",
            function=calc_acc,
        ),
        "roc-auc-macro": MetricRecord(
            name="roc-auc-macro",
            function=calc_roc_auc_ovo,
        ),
        "ap-macro": MetricRecord(
            name="ap-macro",
            function=calc_average_precision,
        ),
        "f1-macro": MetricRecord(
            name="f1-macro",
            function=calc_f1_score_macro,
        ),
        "precision-macro": MetricRecord(
            name="precision-macro",
            function=calc_precision_macro,
        ),
        "recall-macro": MetricRecord(
            name="recall-macro",
            function=calc_recall_macro,
        ),
        "cohen-kappa": MetricRecord(
            name="cohen-kappa",
            function=calc_cohen_kappa,
        ),
    }

    con_metrics: Dict[str, MetricRecord] = {
        "rmse": MetricRecord(
            name="rmse",
            function=partial(calc_rmse, target_transformers=None),
            minimize_goal=True,
        ),
        "r2": MetricRecord(
            name="r2",
            function=calc_r2,
        ),
        "pcc": MetricRecord(
            name="pcc",
            function=calc_pcc,
        ),
        "mae": MetricRecord(
            name="mae",
            function=calc_mae,
            minimize_goal=True,
        ),
        "mape": MetricRecord(
            name="mape",
            function=calc_mape,
            minimize_goal=True,
        ),
        "explained-variance": MetricRecord(
            name="explained-variance",
            function=calc_explained_variance,
        ),
    }

    return cat_metrics, con_metrics


def get_available_survival_metrics(
    target_transformers: dict[str, "al_label_transformers"]
) -> tuple[MetricRecord]:
    survival_metrics: tuple[MetricRecord] = (
        MetricRecord(
            name="c-index",
            function=partial(
                calc_survival_c_index, target_transformers=target_transformers
            ),
            output_type="survival",
            minimize_goal=False,
        ),
    )

    return survival_metrics


def get_default_metrics(
    target_transformers: dict[str, "al_label_transformers"],
    cat_metrics: al_cat_metric_choices,
    con_metrics: al_con_metric_choices,
    cat_averaging_metrics: Optional[al_cat_metric_choices],
    con_averaging_metrics: Optional[al_con_metric_choices],
    output_configs: Sequence[OutputConfig],
) -> "al_metric_record_dict":
    available_cat_metrics, available_con_metrics = get_available_supervised_metrics()

    cat_metric_records = tuple(
        available_cat_metrics[metric]
        for metric in cat_metrics
        if metric in available_cat_metrics
    )

    con_metric_records = tuple(
        available_con_metrics[metric]
        for metric in con_metrics
        if metric in available_con_metrics
    )

    for metric in con_metric_records:
        if metric.name == "rmse":
            metric.function = partial(
                calc_rmse, target_transformers=target_transformers
            )

    cat_for_avg, con_for_avg = parse_averaging_metrics(
        cat_averaging_metrics=cat_averaging_metrics,
        con_averaging_metrics=con_averaging_metrics,
    )
    averaging_functions = get_performance_averaging_functions(
        cat_metric_names=cat_for_avg,
        con_metric_names=con_for_avg,
    )

    surv_metric_records = get_available_survival_metrics(
        target_transformers=target_transformers
    )

    general_metric_info = _build_general_metric_info(
        cat_metric_records=cat_metric_records,
        con_metric_records=con_metric_records,
        surv_metric_records=surv_metric_records,
        output_configs=output_configs,
    )

    default_metrics: al_metric_record_dict = {
        "cat": cat_metric_records,
        "con": con_metric_records,
        "survival": surv_metric_records,
        "averaging_functions": averaging_functions,
        "general_metric_info": general_metric_info,
    }
    return default_metrics


@dataclass
class GeneralMetricInfo:
    all_are_val_only: bool
    base_metric_structure: dict[str, dict[str, dict]]


def _build_general_metric_info(
    cat_metric_records: tuple[MetricRecord, ...],
    con_metric_records: tuple[MetricRecord, ...],
    surv_metric_records: tuple[MetricRecord, ...],
    output_configs: Sequence[OutputConfig],
) -> GeneralMetricInfo:
    """
    This function and the created object is used in in the training loop to:

        1. Determine if all metrics are validation-only metrics. If that is the case
        the metric calculation for training batches is completely skipped.
        2. If (1) is True, we create a base structure for the metric calculation
        which is used in other parts of the code (e.g. for running average metrics).
        The structure is bootstrapped here, even though it is never filled with metrics
        as the general output_name -> target name nested dict structure is e.g.
        used when adding losses to the state dict in the training loop.
    """
    all_cat_are_val = all(metric.only_val for metric in cat_metric_records)
    all_con_are_val = all(metric.only_val for metric in con_metric_records)
    all_surv_are_val = all(metric.only_val for metric in surv_metric_records)
    all_are_val_only = all([all_cat_are_val, all_con_are_val, all_surv_are_val])

    if all_are_val_only:
        logger.debug(
            "All metrics are validation-only metrics, will skip all metric "
            "calculation for training batches."
        )

    base_structure: dict[str, Any] = {}
    for output_config in output_configs:
        output_name = output_config.output_info.output_name
        base_structure[output_name] = {}

        oti = output_config.output_type_info
        match oti:
            case TabularOutputTypeConfig():
                targets = list(oti.target_cat_columns) + list(oti.target_con_columns)
            case SurvivalOutputTypeConfig():
                targets = [oti.event_column]
            case _:
                targets = [output_name]

        for target in targets:
            base_structure[output_name][target] = {}

    return GeneralMetricInfo(
        all_are_val_only=all_are_val_only,
        base_metric_structure=base_structure,
    )


def parse_averaging_metrics(
    cat_averaging_metrics: Optional[al_cat_metric_choices],
    con_averaging_metrics: Optional[al_con_metric_choices],
) -> tuple[al_cat_metric_choices, al_con_metric_choices]:
    cat_parsed, con_parsed = _get_default_averaging_metrics()

    if cat_averaging_metrics:
        _validate_metrics(
            passed_in_metrics=cat_averaging_metrics,
            expected_metrics=[
                "loss",
                "acc",
                "mcc",
                "roc-auc-macro",
                "ap-macro",
            ],
            target_type="categorical",
        )
        cat_parsed = cat_averaging_metrics
    if con_averaging_metrics:
        _validate_metrics(
            passed_in_metrics=con_averaging_metrics,
            expected_metrics=[
                "loss",
                "rmse",
                "pcc",
                "r2",
            ],
            target_type="continuous",
        )
        con_parsed = con_averaging_metrics

    return cat_parsed, con_parsed


def _validate_metrics(
    passed_in_metrics: Sequence[str], expected_metrics: Sequence[str], target_type: str
) -> None:
    for metric in passed_in_metrics:
        if metric not in expected_metrics:
            raise ValueError(
                f"Metric {metric} not found in expected metrics {expected_metrics} for "
                f" {target_type} targets."
            )


def _get_default_averaging_metrics() -> (
    tuple[al_cat_metric_choices, al_con_metric_choices]
):

    cat_names: list[Literal["mcc", "roc-auc-macro", "ap-macro", "acc"]]
    con_names: list[Literal["loss", "pcc", "r2"]]

    cat_names = ["mcc", "roc-auc-macro", "ap-macro"]
    con_names = ["loss", "pcc", "r2"]

    return cat_names, con_names


def get_performance_averaging_functions(
    cat_metric_names: al_cat_metric_choices,
    con_metric_names: al_con_metric_choices,
) -> al_averaging_functions_dict:
    """
    Note that we have the mean(values) else 0.0 to account for some values not being
    computed on the training batches, e.g. ROC-AUC, due some metrics possibly
    raising errors if e.g. there are only negative labels in a batch.
    """

    parsed_con_names = []
    for metric in con_metric_names:
        if metric in ["rmse", "mae", "mape"]:
            logger.warning(
                "Using metric %s for performance averaging, which is affected by "
                "the scale of the target. This can lead to the metric dominating "
                "the performance calculation. Consider using a different metric "
                "for performance averaging.",
            )

        if metric in ["loss", "mae", "rmse", "mape"]:
            parsed_con_names.append(f"1.0-{metric.upper()}")
        else:
            parsed_con_names.append(metric.upper())

    logger.info(
        "Tabular output performance averaging functions across tasks set to averages "
        "of %s for categorical targets and %s for continuous targets. These "
        "values are used to determine overall performance (using the validation set), "
        "which is used to control factors such as early stopping and LR scheduling. "
        "Other output cases (e.g. sequence generation, image generation) "
        "use 1.0-LOSS by default.",
        [i.upper() for i in cat_metric_names],
        parsed_con_names,
    )

    def _calc_cat_averaging_value(
        metric_dict: "al_step_metric_dict",
        output_name: str,
        column_name: str,
        metric_names: al_cat_metric_choices | all_survival_metric_choices,
    ) -> float:
        values = []
        for metric_name in metric_names:
            combined_key = f"{output_name}_{column_name}_{metric_name}"
            value = metric_dict[output_name][column_name].get(combined_key, None)

            if value is None:
                continue

            values.append(float(value))

        return mean(values) if values else 0.0

    def _calc_con_averaging_value(
        metric_dict: "al_step_metric_dict",
        output_name: str,
        column_name: str,
        metric_names: al_con_metric_choices,
    ) -> float:
        values = []
        for metric_name in metric_names:
            combined_key = f"{output_name}_{column_name}_{metric_name}"
            value = metric_dict[output_name][column_name].get(combined_key, None)

            if value is None:
                continue

            if metric_name in ["loss", "rmse", "mae", "mape"]:
                value = 1.0 - value

            values.append(float(value))

        return mean(values) if values else 0.0

    surv_metric_names: all_survival_metric_choices = ["c-index"]

    performance_averaging_functions: al_averaging_functions_dict = {
        "cat": partial(_calc_cat_averaging_value, metric_names=cat_metric_names),
        "con": partial(_calc_con_averaging_value, metric_names=con_metric_names),
        "survival": partial(_calc_cat_averaging_value, metric_names=surv_metric_names),
        "general": partial(_calc_con_averaging_value, metric_names=["loss"]),
    }

    return performance_averaging_functions


@dataclass
class FilteredOutputsAndLabels:
    model_outputs: dict[str, dict[str, torch.Tensor]]
    target_labels: dict[str, dict[str, torch.Tensor]]
    ids: Optional[dict[str, dict[str, list[str]]]]
    common_ids: Optional[list[str]]


def filter_missing_outputs_and_labels(
    batch_ids: list[str],
    model_outputs: dict[str, dict[str, torch.Tensor]],
    target_labels: dict[str, dict[str, torch.Tensor]],
    missing_ids_info: MissingTargetsInfo,
    with_labels: bool = True,
) -> FilteredOutputsAndLabels:

    if missing_ids_info.all_have_same_set:
        return FilteredOutputsAndLabels(
            model_outputs=model_outputs,
            target_labels=target_labels,
            ids=None,
            common_ids=batch_ids,
        )

    filtered_outputs = {}
    filtered_labels = {}
    filtered_ids: dict[str, dict[str, list[str]]] = {}

    missing_per_modality = missing_ids_info.missing_ids_per_modality

    for output_name, output_inner_dict in model_outputs.items():
        cur_output_missing = missing_per_modality.get(output_name, set())
        if not cur_output_missing:
            filtered_outputs[output_name] = output_inner_dict
            filtered_labels[output_name] = (
                target_labels[output_name]
                if with_labels
                else {k: torch.tensor([]) for k in output_inner_dict}
            )
            filtered_ids[output_name] = {k: batch_ids for k in output_inner_dict}
            continue

        missing_ids = missing_ids_info.missing_ids_per_modality[output_name]
        valid_indices = [i for i, id_ in enumerate(batch_ids) if id_ not in missing_ids]

        device = next(iter(output_inner_dict.values())).device
        valid_indices_tensor = torch.tensor(
            valid_indices,
            device=device,
            dtype=torch.long,
        )

        filtered_outputs[output_name] = {
            k: v.index_select(0, valid_indices_tensor)
            for k, v in output_inner_dict.items()
        }

        if with_labels:
            filtered_labels[output_name] = {
                k: v.index_select(0, valid_indices_tensor)
                for k, v in target_labels[output_name].items()
            }
        else:
            filtered_labels[output_name] = {
                k: torch.tensor([]) for k in output_inner_dict
            }
        filtered_ids[output_name] = {
            k: [batch_ids[i] for i in valid_indices] for k in output_inner_dict
        }

    return FilteredOutputsAndLabels(
        model_outputs=filtered_outputs,
        target_labels=filtered_labels,
        ids=filtered_ids,
        common_ids=None,
    )


@dataclass
class FilteredTabularTargets:
    model_outputs: torch.Tensor
    target_labels: torch.Tensor
    ids: list[str]


def filter_tabular_missing_targets(
    outputs: torch.Tensor,
    target_labels: torch.Tensor,
    ids: list[str],
    target_type: str,
) -> FilteredTabularTargets:
    nan_labels_mask = torch.isnan(target_labels)
    if target_type == "cat":
        nan_labels_mask = torch.logical_or(nan_labels_mask, target_labels == -1)

    if ids:
        cur_ids_np = np.array(ids)
        nan_labels_mask_np = nan_labels_mask.cpu().numpy()
        ids = cur_ids_np[~nan_labels_mask_np].tolist()

    predictions = outputs[~nan_labels_mask]
    target_labels = target_labels[~nan_labels_mask]

    return FilteredTabularTargets(
        model_outputs=predictions,
        target_labels=target_labels,
        ids=ids,
    )


@dataclass
class FilteredSurvivalTargets:
    model_outputs: torch.Tensor
    events: torch.Tensor
    times: torch.Tensor
    ids: list[str]


def filter_survival_missing_targets(
    model_outputs: torch.Tensor,
    events: torch.Tensor,
    times: torch.Tensor,
    cur_ids: list[str],
) -> FilteredSurvivalTargets:
    nan_labels_mask = events == -1
    cur_times_nan = torch.isnan(times)
    nan_labels_mask = torch.logical_or(nan_labels_mask, cur_times_nan)

    predictions = model_outputs[~nan_labels_mask]
    events = events[~nan_labels_mask]
    times = times[~nan_labels_mask]

    if cur_ids:
        nan_labels_mask_np = nan_labels_mask.cpu().numpy()
        cur_ids = np.array(cur_ids)[~nan_labels_mask_np].tolist()

    return FilteredSurvivalTargets(
        model_outputs=predictions,
        events=events,
        times=times,
        ids=cur_ids,
    )


def general_torch_to_numpy(tensor: torch.Tensor) -> np.ndarray:
    array = tensor.cpu().to(dtype=torch.float32).detach().numpy()
    return array



from collections import Counter
from statistics import mean
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Sequence, Tuple

import numpy as np
import torch
from torch.utils.data import WeightedRandomSampler

from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.datasets import DatasetBase, HybridStorage  # noqa: F401

logger = get_logger(name=__name__, tqdm_compatible=True)

al_sample_weight_and_counts = Dict[str, torch.Tensor | list[int]]


def get_weighted_random_sampler(
    target_storage: "HybridStorage",
    columns_to_sample: List[str],
) -> WeightedRandomSampler:
    """
    Takes a HybridStorage with target data and returns a WeightedRandomSampler
    based on the specified columns.
    """
    parsed_weighted_sample_columns = _build_weighted_sample_dict_from_config_sequence(
        config_list=columns_to_sample
    )

    all_column_weights = {}
    logger.debug("Setting up weighted sampling statistics.")

    for output_name, weighted_columns_list in parsed_weighted_sample_columns.items():
        logger.debug(f"Setting up weighted sampling for output '{output_name}'.")
        cur_column_weights = _gather_column_sampling_weights(
            target_storage=target_storage,
            output_name=output_name,
            columns_to_sample=weighted_columns_list,
        )
        for cur_target, cur_weight_object in cur_column_weights.items():
            all_column_weights[f"{output_name}.{cur_target}"] = cur_weight_object

    samples_weighted, num_sample_per_epoch = _aggregate_column_sampling_weights(
        all_target_columns_weights_and_counts=all_column_weights
    )

    logger.debug(
        "Num samples per epoch according to average target class counts in %s: %d",
        columns_to_sample,
        num_sample_per_epoch,
    )

    return WeightedRandomSampler(
        weights=samples_weighted,
        num_samples=num_sample_per_epoch,
        replacement=True,
    )


def _build_weighted_sample_dict_from_config_sequence(
    config_list: List[str],
) -> Dict[str, List[str]]:
    weighted_sample_dict = {}

    for weighted_sample_config_string in config_list:
        if weighted_sample_config_string == "all":
            return {"all": ["all"]}

        output_name, _ = weighted_sample_config_string.split("__", 1)
        sample_column = weighted_sample_config_string
        if output_name not in weighted_sample_dict:
            weighted_sample_dict[output_name] = [sample_column]
        else:
            weighted_sample_dict[output_name].append(sample_column)

    return weighted_sample_dict


def _gather_column_sampling_weights(
    target_storage: "HybridStorage",
    output_name: str,
    columns_to_sample: Iterable[str],
) -> Dict[str, al_sample_weight_and_counts]:
    all_target_label_weight_dicts: dict[str, al_sample_weight_and_counts] = {}

    for column in columns_to_sample:
        try:
            if column == "all":
                continue

            if any(col.name == column for col in target_storage.numeric_int_columns):
                col_idx = next(
                    i
                    for i, col in enumerate(target_storage.numeric_int_columns)
                    if col.name == column
                )
                assert target_storage.numeric_int_data is not None
                cur_label_data = target_storage.numeric_int_data[col_idx].numpy()

            elif any(
                col.name == column for col in target_storage.numeric_float_columns
            ):
                logger.warning(
                    f"Computing sampling weights for float column {column}. "
                    "This is unusual as sampling weights are typically used for "
                    "categorical data."
                )

                col_idx = next(
                    i
                    for i, col in enumerate(target_storage.numeric_float_columns)
                    if col.name == column
                )
                assert target_storage.numeric_float_data is not None
                cur_label_data = target_storage.numeric_float_data[col_idx].numpy()

            else:
                raise KeyError(f"Column {column} not found in numeric data")

            cur_label_iterable_int = (
                int(i) if not np.isnan(i) else np.nan for i in cur_label_data
            )

            cur_weight_dict = _get_column_label_weights_and_counts(
                label_iterable=cur_label_iterable_int,
                column_name=column,
            )

            logger.debug(
                "Label counts in column %s: %s", column, cur_weight_dict["label_counts"]
            )

            all_target_label_weight_dicts[column] = cur_weight_dict

        except ValueError as e:
            logger.warning(f"Skipping column {column} due to error: {str(e)}.")
        except KeyError:
            logger.warning(f"Skipping column {column} due to missing data.")

    if not all_target_label_weight_dicts:
        logger.warning("No valid columns found for sampling weights")

    return all_target_label_weight_dicts


def _get_column_label_weights_and_counts(
    label_iterable: Iterable[int | float], column_name: Optional[str] = None
) -> dict[str, torch.Tensor | list[int]]:
    def _check_labels(label_list: List[int]):
        labels_set = set(label_list)
        found_labels = sorted(list(labels_set))
        expected_labels = list(range(len(found_labels)))

        if found_labels != expected_labels:
            raise ValueError(
                f"When setting up weighed sampling for column {column_name}, "
                f"all labels must be present in the training set. "
                f"Expected at least {max(expected_labels)} labels, "
                f"but got {len(found_labels)}. "
                "This is likely due to a mismatch between the training set and the "
                "validation set, possibly due to rare labels in the data that e.g. "
                "only appear in the validation set after splitting. "
                "Weighed sampling is therefore not supported for this "
                "column."
            )

    labels = np.array(list(label_iterable))

    valid_mask = ~np.isnan(labels)
    valid_labels = labels[valid_mask].astype(int)

    _check_labels(label_list=valid_labels.tolist())

    label_counts = [int(i[1]) for i in sorted(Counter(valid_labels).items())]

    weights = 1.0 / torch.tensor(label_counts, dtype=torch.float32)

    samples_weighted = torch.full((len(labels),), float("nan"), dtype=torch.float32)
    samples_weighted[valid_mask] = weights[valid_labels]

    output_dict: dict[str, torch.Tensor | list[int]] = {
        "samples_weighted": samples_weighted,
        "label_counts": label_counts,
    }
    return output_dict


def _aggregate_column_sampling_weights(
    all_target_columns_weights_and_counts: dict[str, al_sample_weight_and_counts]
) -> Tuple[Sequence[float], int]:
    """
    We sum up the normalized weights for each target column to create the final sampling
    weights.

    As for the samples per epoch, we take the average of the class counts per target,
    then sum those up.
    """
    all_weights = torch.stack(
        [
            torch.Tensor(i["samples_weighted"])
            for i in all_target_columns_weights_and_counts.values()
        ],
        dim=1,
    )

    all_weights = torch.nan_to_num(all_weights, nan=0.0)
    all_weights_summed = all_weights.sum(dim=1)

    samples_per_epoch = int(
        mean(
            mean(i["label_counts"])
            for i in all_target_columns_weights_and_counts.values()
        )
    )
    samples_per_epoch = min(len(all_weights_summed), samples_per_epoch)

    all_weights_summed_list = all_weights_summed.tolist()

    return all_weights_summed_list, samples_per_epoch



import json
from dataclasses import asdict
from pathlib import Path
from typing import Any, Callable, Protocol, cast

import numpy as np
import torch
import yaml
from aislib.misc_utils import ensure_path_exists

from eir import __version__
from eir.setup.config_setup_modules.config_setup_utils import object_to_primitives
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


def encode_numpy(obj: np.ndarray) -> dict[str, Any]:
    return {"__np_array__": obj.tolist(), "dtype": str(obj.dtype)}


def encode_torch(obj: torch.Tensor) -> dict[str, Any]:
    return {
        "__torch_tensor__": obj.cpu().numpy().tolist(),
        "dtype": str(obj.dtype),
        "device": str(obj.device),
    }


def encode_tuple(obj: tuple) -> list:
    return list(obj)


def encode_primitive(obj: Any) -> Any:
    return obj


class EncoderProtocol(Protocol):
    def __call__(self, obj: Any) -> Any: ...


def get_encoder(obj: Any) -> EncoderProtocol:
    encoders = {
        np.ndarray: encode_numpy,
        torch.Tensor: encode_torch,
        tuple: encode_tuple,
        int: encode_primitive,
        float: encode_primitive,
        str: encode_primitive,
        bool: encode_primitive,
        type(None): encode_primitive,
    }

    encoder = encoders.get(type(obj), encode_primitive)

    return cast(EncoderProtocol, encoder)


def custom_encoder(obj: Any) -> Any:
    return get_encoder(obj=obj)(obj=obj)


def decode_numpy(dct: dict[str, Any]) -> np.ndarray:
    return np.array(dct["__np_array__"], dtype=np.dtype(dct["dtype"]))


def decode_torch(dct: dict[str, Any]) -> torch.Tensor:
    tensor = torch.tensor(
        dct["__torch_tensor__"], dtype=getattr(torch, str(dct["dtype"]).split(".")[-1])
    )
    return tensor.to(dct["device"])


def decode_primitive(dct: Any) -> Any:
    return dct


def get_decoder(dct: Any) -> Callable:
    if isinstance(dct, dict):
        decoders = {
            "__np_array__": decode_numpy,
            "__torch_tensor__": decode_torch,
        }
        for key in decoders:
            if key in dct:
                return decoders[key]
    return decode_primitive


def custom_decoder(dct: Any) -> Any:
    return get_decoder(dct=dct)(dct=dct)


def serialize_dataclass(obj: Any) -> str:
    return json.dumps(obj=asdict(obj), default=custom_encoder, indent=2)


def deserialize_dataclass(cls: type, json_str: str) -> Any:
    data = json.loads(s=json_str, object_hook=custom_decoder)
    return cls(**data)


def save_dataclass(obj: Any, file_path: Path) -> None:
    file_path.write_text(data=serialize_dataclass(obj=obj))


def load_dataclass(cls: type, file_path: Path) -> Any:
    return deserialize_dataclass(cls=cls, json_str=file_path.read_text())


def get_run_folder_from_model_path(model_path: str) -> Path:
    model_path_object = Path(model_path)
    assert model_path_object.exists()

    run_folder = model_path_object.parents[1]
    assert run_folder.exists()

    return run_folder


def dump_config_to_yaml(
    config: Any,
    output_path: Path,
    stripped: bool = True,
) -> None:
    object_primitive = object_to_primitives(obj=config)

    if stripped:
        object_primitive = strip_config(config=object_primitive)

    ensure_path_exists(path=output_path, is_folder=False)
    with open(output_path, "w") as outfile:
        yaml.dump(data=object_primitive, stream=outfile)


def check_version(run_folder: Path) -> None:
    version_file = run_folder / "meta/eir_version.txt"
    if not version_file.exists():
        return

    cur_version = __version__
    with open(version_file, "r") as f:
        loaded_version = f.read().strip()

    if cur_version != loaded_version:
        logger.warning(
            f"The version of EIR used to train this model is {loaded_version}, "
            f"while the current version is {cur_version}. "
            f"This may cause unexpected behavior and subtle bugs."
        )


def strip_config(config: dict | list[dict]) -> dict | list[dict]:
    keys_to_strip = [
        "basic_experiment.manual_valid_ids_file",
        "basic_experiment.output_folder",
        "input_info.input_source",
        "output_info.output_source",
        "input_type_info.vocab_file",
        "input_type_info.snp_file",
        "input_type_info.subset_snps_file",
        "output_type_info.vocab_file",
    ]
    replacements = {k: None for k in keys_to_strip}

    if isinstance(config, list):
        stripped_configs = []

        for cur_config in config:
            cur_config_stripped = replace_dict_values(
                target=cur_config,
                replacements=replacements,
            )
            stripped_configs.append(cur_config_stripped)

        return stripped_configs

    config_stripped = replace_dict_values(
        target=config,
        replacements=replacements,
    )

    return config_stripped


def replace_dict_values(
    target: dict[str, Any],
    replacements: dict[str, Any],
) -> dict[str, Any]:

    def recursive_replace(
        d: dict[str, Any],
        path: str = "",
    ) -> None:
        for key, value in d.items():
            current_path = f"{path}.{key}" if path else key
            if current_path in replacements:
                d[key] = replacements[current_path]
            elif isinstance(value, dict):
                recursive_replace(d=value, path=current_path)

    result = target.copy()
    recursive_replace(d=result)
    return result



from typing import TYPE_CHECKING, Any, Dict, Type, Union

from eir.models.input.array.array_models import (
    ArrayModelConfig,
    get_array_config_dataclass_mapping,
)
from eir.models.input.image.image_models import ImageModelConfig
from eir.models.input.omics.omics_models import (
    OmicsModelConfig,
    get_omics_config_dataclass_mapping,
)
from eir.models.input.sequence.transformer_models import (
    BasicTransformerFeatureExtractorModelConfig,
    SequenceModelConfig,
)
from eir.models.input.tabular.tabular import (
    SimpleTabularModelConfig,
    TabularModelConfig,
)
from eir.setup import schemas
from eir.setup.config_setup_modules.config_setup_utils import (
    validate_keys_against_dataclass,
)
from eir.setup.tensor_broker_setup import set_up_tensor_broker_config
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


if TYPE_CHECKING:
    from eir.setup.config import al_input_types


def init_input_config(yaml_config_as_dict: Dict[str, Any]) -> schemas.InputConfig:
    cfg = yaml_config_as_dict

    validate_keys_against_dataclass(
        input_dict=cfg,
        dataclass_type=schemas.InputConfig,
        name=cfg.get("input_info", {}).get("input_name", ""),
    )

    input_info_object = schemas.InputDataConfig(**cfg["input_info"])

    input_schema_map = get_inputs_schema_map()
    input_type_info_class = input_schema_map[input_info_object.input_type]
    input_type_info_kwargs = cfg.get("input_type_info", {})
    input_type_info_object = input_type_info_class(**input_type_info_kwargs)

    _validate_input_type_info_object(
        input_type_info_object=input_type_info_object,
        input_source=input_info_object.input_source,
    )

    model_config = set_up_input_feature_extractor_config(
        input_info_object=input_info_object,
        input_type_info_object=input_type_info_object,
        model_init_kwargs_base=cfg.get("model_config", {}),
    )

    pretrained_config = set_up_pretrained_config(
        pretrained_config_dict=cfg.get("pretrained_config", None)
    )

    interpretation_config = set_up_interpretation_config(
        input_type=input_info_object.input_type,
        interpretation_config_dict=cfg.get("interpretation_config", None),
    )

    tensor_broker_config = set_up_tensor_broker_config(
        tensor_broker_config=cfg.get("tensor_broker_config", {})
    )

    input_config = schemas.InputConfig(
        input_info=input_info_object,
        input_type_info=input_type_info_object,
        pretrained_config=pretrained_config,
        model_config=model_config,
        interpretation_config=interpretation_config,
        tensor_broker_config=tensor_broker_config,
    )

    return input_config


def get_inputs_schema_map() -> Dict[
    str,
    Union[
        Type[schemas.OmicsInputDataConfig],
        Type[schemas.TabularInputDataConfig],
        Type[schemas.SequenceInputDataConfig],
        Type[schemas.ByteInputDataConfig],
    ],
]:
    mapping = {
        "omics": schemas.OmicsInputDataConfig,
        "tabular": schemas.TabularInputDataConfig,
        "sequence": schemas.SequenceInputDataConfig,
        "bytes": schemas.ByteInputDataConfig,
        "image": schemas.ImageInputDataConfig,
        "array": schemas.ArrayInputDataConfig,
    }

    return mapping


def _validate_input_type_info_object(
    input_type_info_object: "al_input_types",
    input_source: str,
) -> None:
    ito = input_type_info_object
    match ito:
        case schemas.TabularInputDataConfig():
            con = ito.input_con_columns
            cat = ito.input_cat_columns
            common = set(con).intersection(set(cat))
            if common:
                raise ValueError(
                    f"Found columns passed in both continuous and categorical inputs: "
                    f"{common}. In input source: {input_source}."
                    f"Please make sure that each column is only in one of the two, "
                    f"or create differently named copies of the columns."
                )


def set_up_input_feature_extractor_config(
    input_info_object: schemas.InputDataConfig,
    input_type_info_object: "al_input_types",
    model_init_kwargs_base: dict,
) -> schemas.al_feature_extractor_configs:
    input_type = input_info_object.input_type

    model_config_class = get_input_feature_extractor_config_class(input_type=input_type)

    model_type = model_init_kwargs_base.get("model_type", None)
    if not model_type:
        try:
            model_type = getattr(model_config_class, "model_type")
        except AttributeError:
            raise AttributeError(
                "Not model type specified in model config and could not find default "
                "value for '%s'.",
                input_type,
            )

        logger.info(
            "Input model type not specified in model configuration for input name "
            "'%s', attempting to grab default value.",
            input_info_object.input_name,
        )

    model_type_init_config = set_up_feature_extractor_init_config(
        input_info_object=input_info_object,
        input_type_info_object=input_type_info_object,
        model_init_kwargs=model_init_kwargs_base.get("model_init_config", {}),
        model_type=model_type,
    )

    common_kwargs = {
        "model_type": model_type,
        "model_init_config": model_type_init_config,
    }
    other_specific_kwargs = {
        k: v for k, v in model_init_kwargs_base.items() if k not in common_kwargs
    }
    model_config_kwargs = {**common_kwargs, **other_specific_kwargs}
    model_config = model_config_class(**model_config_kwargs)

    return model_config


def get_input_feature_extractor_config_class(
    input_type: str,
) -> schemas.al_feature_extractor_configs_classes:
    model_config_setup_map = get_input_feature_extractor_config_init_class_map()

    return model_config_setup_map[input_type]


def get_input_feature_extractor_config_init_class_map() -> (
    Dict[str, schemas.al_feature_extractor_configs_classes]
):
    mapping = {
        "tabular": TabularModelConfig,
        "omics": OmicsModelConfig,
        "sequence": SequenceModelConfig,
        "bytes": SequenceModelConfig,
        "image": ImageModelConfig,
        "array": ArrayModelConfig,
    }

    return mapping


def set_up_feature_extractor_init_config(
    input_info_object: schemas.InputDataConfig,
    input_type_info_object: "al_input_types",
    model_init_kwargs: dict,
    model_type: str,
) -> Dict:
    if getattr(input_type_info_object, "pretrained_model", None):
        return {}

    not_from_eir = get_is_not_eir_model_condition(
        input_info_object=input_info_object,
        model_type=model_type,
    )
    if not_from_eir:
        return model_init_kwargs

    if not model_init_kwargs:
        model_init_kwargs = {}

    model_config_map = get_feature_extractor_config_type_init_callable_map()
    model_config_callable = model_config_map[model_type]

    model_config = model_config_callable(**model_init_kwargs)

    return model_config


def get_is_not_eir_model_condition(
    input_info_object: schemas.InputDataConfig, model_type: str
) -> bool:
    is_possibly_external = getattr(input_info_object, "input_type") in (
        "sequence",
        "bytes",
        "image",
    )
    is_unknown_model = model_type not in ("sequence-default", "cnn", "lcl")
    not_from_eir = is_possibly_external and is_unknown_model
    return not_from_eir


def get_feature_extractor_config_type_init_callable_map() -> Dict[str, Type]:
    omics_mapping = get_omics_config_dataclass_mapping()
    array_mapping = get_array_config_dataclass_mapping()
    other_mapping = {
        "tabular": SimpleTabularModelConfig,
        "sequence-default": BasicTransformerFeatureExtractorModelConfig,
    }

    mapping = omics_mapping | array_mapping | other_mapping
    return mapping


def set_up_pretrained_config(
    pretrained_config_dict: Union[None, Dict[str, Any]]
) -> Union[None, schemas.BasicPretrainedConfig]:
    if pretrained_config_dict is None:
        return None

    config_class = get_pretrained_config_class()
    if config_class is None:
        return None

    config_object = config_class(**pretrained_config_dict)

    return config_object


def get_pretrained_config_class() -> Type[schemas.BasicPretrainedConfig]:
    return schemas.BasicPretrainedConfig


def set_up_interpretation_config(
    input_type: str, interpretation_config_dict: Union[None, Dict[str, Any]]
) -> Union[None, schemas.BasicInterpretationConfig]:
    config_class = get_interpretation_config_class(input_type=input_type)
    if config_class is None:
        return None

    if interpretation_config_dict is None:
        interpretation_config_dict = {}

    config_object = config_class(**interpretation_config_dict)

    return config_object


def get_interpretation_config_class(
    input_type: str,
) -> Union[None, Type[schemas.BasicInterpretationConfig]]:
    mapping = get_interpretation_config_schema_map()

    return mapping.get(input_type, None)


def get_interpretation_config_schema_map() -> (
    Dict[str, Type[schemas.BasicInterpretationConfig]]
):
    mapping = {
        "sequence": schemas.BasicInterpretationConfig,
        "image": schemas.BasicInterpretationConfig,
    }

    return mapping



import ast
import json
import operator
from collections import defaultdict
from dataclasses import fields, is_dataclass
from functools import reduce
from pathlib import Path
from typing import (
    Any,
    DefaultDict,
    Dict,
    Generator,
    Iterable,
    MutableMapping,
    Optional,
    Type,
)

import yaml

from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


def get_yaml_iterator_with_injections(
    yaml_config_files: Iterable[str], extra_cl_args: Optional[list[str]]
) -> Generator[Dict, None, None]:
    if not extra_cl_args:
        yield from get_yaml_to_dict_iterator(yaml_config_files=yaml_config_files)
        return

    for yaml_config_file in yaml_config_files:
        loaded_yaml = load_yaml_config(config_path=yaml_config_file)

        yaml_file_path_object = Path(yaml_config_file)
        for extra_arg in extra_cl_args:
            extra_arg_parsed = extra_arg.lstrip("--")
            target_file, str_to_inject = extra_arg_parsed.split(".", 1)

            if target_file == yaml_file_path_object.stem:
                dict_to_inject = convert_cl_str_to_dict(str_=str_to_inject)

                logger.debug("Injecting %s into %s", dict_to_inject, loaded_yaml)
                loaded_yaml = recursive_dict_inject(
                    dict_=loaded_yaml, dict_to_inject=dict_to_inject
                )

        yield loaded_yaml


def convert_cl_str_to_dict(str_: str) -> dict:
    def _infinite_dict() -> DefaultDict:
        return defaultdict(_infinite_dict)

    infinite_dict = _infinite_dict()

    keys, final_value = str_.split("=", 1)
    keys_split = keys.split(".")

    try:
        final_value_parsed = ast.literal_eval(final_value)
    except (ValueError, SyntaxError):
        final_value_parsed = final_value

    inner_most_dict = reduce(operator.getitem, keys_split[:-1], infinite_dict)
    inner_most_dict[keys_split[-1]] = final_value_parsed

    dict_primitive = object_to_primitives(obj=infinite_dict)
    return dict_primitive


def get_yaml_to_dict_iterator(
    yaml_config_files: Iterable[str],
) -> Generator[dict, None, None]:
    for yaml_config in yaml_config_files:
        yield load_yaml_config(config_path=yaml_config)


def load_yaml_config(config_path: str) -> Dict[str, Any]:
    with open(config_path, "r") as yaml_file:
        config_as_dict = yaml.load(stream=yaml_file, Loader=yaml.FullLoader)

    return config_as_dict


def recursive_dict_inject(
    dict_: MutableMapping,
    dict_to_inject: MutableMapping,
) -> dict:
    for cur_key, cur_value in dict_to_inject.items():
        if cur_key not in dict_:
            dict_[cur_key] = {}

        old_dict_value = dict_.get(cur_key)
        cur_is_dict = isinstance(cur_value, MutableMapping)
        old_is_dict = isinstance(old_dict_value, MutableMapping)
        if cur_is_dict and old_is_dict:
            assert isinstance(cur_value, MutableMapping)
            assert isinstance(old_dict_value, MutableMapping)
            recursive_dict_inject(dict_=old_dict_value, dict_to_inject=cur_value)
        else:
            dict_[cur_key] = cur_value

    return dict(dict_)


def object_to_primitives(obj):
    return json.loads(json.dumps(obj, default=lambda o: o.__dict__))


def validate_keys_against_dataclass(
    input_dict: Dict[str, Any], dataclass_type: Type, name: str = ""
) -> None:
    if not is_dataclass(dataclass_type):
        raise TypeError(f"Provided type {dataclass_type.__name__} is not a dataclass")

    expected_keys = {field_.name for field_ in fields(dataclass_type)}

    actual_keys = set(input_dict.keys())

    unexpected_keys = actual_keys - expected_keys
    if unexpected_keys:
        message = (
            f"Unexpected keys found in configuration: '{', '.join(unexpected_keys)}'. "
            f"Expected keys of type '{dataclass_type.__name__}': "
            f"'{', '.join(expected_keys)}'."
        )

        if name:
            message = f"{name}: {message}"

        raise KeyError(message)



import json
from copy import deepcopy
from dataclasses import dataclass
from itertools import cycle
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Tuple, Union

import numpy as np
import torch
from aislib.misc_utils import ensure_path_exists, get_logger
from PIL import Image
from sklearn.preprocessing import LabelEncoder, StandardScaler
from torch.utils.data import DataLoader, Dataset

from eir.data_load.data_preparation_modules.prepare_array import (
    array_load_wrapper,
    prepare_array_data,
    un_normalize_array,
)
from eir.data_load.data_preparation_modules.prepare_bytes import (
    bytes_load_wrapper,
    prepare_bytes_data,
)
from eir.data_load.data_preparation_modules.prepare_image import (
    image_load_wrapper,
    prepare_image_data,
)
from eir.data_load.data_preparation_modules.prepare_omics import (
    omics_load_wrapper,
    prepare_one_hot_omics_data,
)
from eir.data_load.data_preparation_modules.prepare_sequence import (
    prepare_sequence_data,
)
from eir.data_load.data_streaming.streaming_dataset_utils import (
    streamline_sequence_manual_data,
)
from eir.data_load.data_utils import Batch
from eir.data_load.datasets import al_getitem_return
from eir.data_load.label_setup import (
    al_label_transformers,
    streamline_values_for_transformers,
)
from eir.interpretation.interpret_image import un_normalize_image
from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo
from eir.setup.input_setup_modules.setup_array import (
    ArrayNormalizationStats,
    ComputedArrayInputInfo,
)
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import (
    ComputedImageInputInfo,
    ImageNormalizationStats,
)
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import (
    ComputedSequenceInputInfo,
    TokenizerProtocolPreSplit,
    TokenizerProtocolRaw,
)
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.input_setup_modules.torchtext_port.vocab import Vocab
from eir.setup.schemas import (
    ArrayInputDataConfig,
    ByteInputDataConfig,
    ImageInputDataConfig,
    OmicsInputDataConfig,
    SequenceInputDataConfig,
    TabularInputDataConfig,
)
from eir.train_utils.utils import call_hooks_stage_iterable

if TYPE_CHECKING:
    from eir.predict import PredictExperiment, PredictHooks
    from eir.serve import ServeExperiment
    from eir.setup.input_setup import al_input_objects_as_dict
    from eir.train import Experiment
    from eir.train_utils.step_logic import Hooks

logger = get_logger(name=__name__)


def remove_special_tokens_from_string(
    string: str, special_tokens: "SpecialTokens"
) -> str:
    """
    TODO:   Deprecate in favor of `remove_special_tokens`, due to not being guaranteed
            to handle spaces / split_on correctly, only removes the tokens themselves
            and creates extra spaces.
    """
    token_names = ["mask_token", "pad_token", "eos_token", "bos_token", "unk_token"]
    for token in token_names:
        assert hasattr(special_tokens, token)
        token_value = getattr(special_tokens, token)
        string = string.replace(token_value, "")
    return string


def remove_special_tokens(
    tokens: list[int], special_tokens: "SpecialTokens"
) -> list[int]:
    token_names = ["mask_idx", "pad_idx", "eos_idx", "bos_idx", "unk_idx"]
    for token in token_names:
        assert hasattr(special_tokens, token)
        token_value = getattr(special_tokens, token)
        tokens = [i for i in tokens if i != token_value]
    return tokens


def convert_model_inputs_to_raw(
    inputs_to_model: dict[str, torch.Tensor | dict[str, torch.Tensor]],
    input_objects: "al_input_objects_as_dict",
) -> dict[str, str | np.ndarray | dict[str, np.ndarray] | Image.Image]:
    raw_inputs = {}
    for input_name, data in inputs_to_model.items():
        input_object = input_objects[input_name]

        raw_input: str | np.ndarray | dict[str, np.ndarray] | Image.Image
        match input_object:
            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                assert isinstance(data, dict)
                raw_input = convert_tabular_input_to_raw(
                    data=data,
                    input_transformers=input_object.labels.label_transformers,
                )

            case ComputedOmicsInputInfo() | ComputedBytesInputInfo():
                assert isinstance(data, torch.Tensor)
                raw_input = data.numpy().squeeze()

            case ComputedSequenceInputInfo():
                assert isinstance(data, torch.Tensor)
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, SequenceInputDataConfig)
                assert input_object.tokenizer is not None
                raw_input = decode_tokens(
                    tokens=data.numpy().tolist(),
                    vocab=input_object.vocab,
                    split_on=input_type_info.split_on,
                )
                special_tokens = get_special_tokens(
                    tokenizer=input_object.tokenizer, vocab=input_object.vocab
                )
                raw_input = remove_special_tokens_from_string(
                    string=raw_input,
                    special_tokens=special_tokens,
                )

            case ComputedImageInputInfo():
                assert isinstance(data, torch.Tensor)
                raw_input = convert_image_input_to_raw(
                    data=data, normalization_stats=input_object.normalization_stats
                )

            case ComputedArrayInputInfo():
                assert isinstance(data, torch.Tensor)
                raw_input = convert_array_input_to_raw(
                    data=data, normalization_stats=input_object.normalization_stats
                )

            case _:
                raise NotImplementedError()

        raw_inputs[input_name] = raw_input

    return raw_inputs


def convert_image_input_to_raw(
    data: torch.Tensor, normalization_stats: ImageNormalizationStats
) -> Image.Image:
    data_np: np.ndarray = data.numpy()
    assert data_np.ndim == 3, "Input should be 3D"

    cur_input = un_normalize_image(
        normalized_img=data_np,
        normalization_stats=normalization_stats,
    )

    cur_input_hwc = np.moveaxis(cur_input, 0, -1)
    raw_input_uint = (cur_input_hwc * 255).astype(np.uint8)

    n_channels = raw_input_uint.shape[-1]
    mode: Optional[str]
    match n_channels:
        case 1:
            mode = "L"
            raw_input_uint = raw_input_uint.squeeze(axis=-1)
        case 3:
            mode = "RGB"
        case 4:
            mode = "RGBA"
        case _:
            mode = None

    return Image.fromarray(raw_input_uint, mode=mode)


def convert_tabular_input_to_raw(
    data: Dict[str, torch.Tensor], input_transformers: al_label_transformers
) -> Dict[str, np.ndarray]:
    all_reversed: Dict[str, np.ndarray] = {}
    for col_name, tensor_data in data.items():
        transformer = input_transformers[col_name]
        tensor_data_reshaped = tensor_data.numpy().reshape(-1, 1)
        assert tensor_data_reshaped.shape[0] > 0, "Empty tensor"

        it = transformer.inverse_transform
        if isinstance(transformer, StandardScaler):
            cur_reversed = it(tensor_data_reshaped)
        elif isinstance(transformer, LabelEncoder):
            cur_reversed = it(tensor_data_reshaped.ravel())
        else:
            raise NotImplementedError()

        cur_reversed = cur_reversed.squeeze()
        assert cur_reversed.ndim == 0

        cur_reversed = cur_reversed.item()
        all_reversed[col_name] = cur_reversed
    return all_reversed


def convert_array_input_to_raw(
    data: torch.Tensor, normalization_stats: Optional[ArrayNormalizationStats]
) -> np.ndarray:
    data_un_normalized = un_normalize_array(
        array=data, normalization_stats=normalization_stats
    )

    data_un_normalized_numpy = data_un_normalized.numpy()

    return data_un_normalized_numpy


def general_pre_process_prepared_inputs(
    prepared_inputs: dict[str, Any],
    target_labels: dict[str, Any],
    sample_ids: list[str],
    experiment: Union["Experiment", "PredictExperiment", "ServeExperiment"],
    custom_hooks: Union["Hooks", "PredictHooks"],
) -> Batch:
    """
    The custom hooks here make sure we are doing basic processing,
    i.e. not the sequence output specific things we define here in the train module.
    """

    loader_batch = (prepared_inputs, target_labels, sample_ids)

    batch_prep_hook_kwargs = {"experiment": experiment}
    state = call_hooks_stage_iterable(
        hook_iterable=custom_hooks.step_func_hooks.base_prepare_batch,
        common_kwargs={"loader_batch": loader_batch, **batch_prep_hook_kwargs},
        state=None,
    )
    batch = state["batch"]

    batch_final = Batch(inputs=batch.inputs, target_labels={}, ids=list())

    return batch_final


@dataclass
class SpecialTokens:
    mask_idx: int
    pad_idx: int
    eos_idx: int
    bos_idx: int
    unk_idx: int

    mask_token: str
    pad_token: str
    eos_token: str
    bos_token: str
    unk_token: str


def get_special_tokens(
    tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit, vocab: Vocab
) -> SpecialTokens:
    keys_and_defaults = (
        ("mask_token", "<mask>"),
        ("pad_token", "<pad>"),
        ("bos_token", "<bos>"),
        ("eos_token", "<eos>"),
        ("unk_token", "<unk>"),
    )

    token_values = {}
    index_values = {}

    for key, default in keys_and_defaults:
        cur_token = getattr(tokenizer, key, default)
        token_values[key] = cur_token

        cur_index = vocab[cur_token]
        idx_kwarg_key = key.replace("_token", "_idx")
        index_values[idx_kwarg_key] = cur_index

    special_tokens = SpecialTokens(
        mask_idx=index_values["mask_idx"],
        pad_idx=index_values["pad_idx"],
        eos_idx=index_values["eos_idx"],
        bos_idx=index_values["bos_idx"],
        unk_idx=index_values["unk_idx"],
        mask_token=token_values["mask_token"],
        pad_token=token_values["pad_token"],
        eos_token=token_values["eos_token"],
        bos_token=token_values["bos_token"],
        unk_token=token_values["unk_token"],
    )

    return special_tokens


def decode_tokens(
    tokens: list[int],
    vocab: Vocab,
    split_on: str | None,
) -> str:
    tokens_decoded = vocab.lookup_tokens(indices=tokens)
    if split_on is not None:
        generated_sample = split_on.join(tokens_decoded)
    else:
        generated_sample = "".join(tokens_decoded)

    return generated_sample


def _streamline_tabular_data_for_transformers(
    tabular_input: dict[str, np.ndarray], transformers: al_label_transformers
) -> dict[str, torch.Tensor]:
    parsed_output = {}
    for name, value in tabular_input.items():
        cur_transformer = transformers[name]
        value_np = np.array([value])
        value_streamlined = streamline_values_for_transformers(
            transformer=cur_transformer, values=value_np
        )
        value_transformed = cur_transformer.transform(value_streamlined)
        value_tensor = torch.from_numpy(value_transformed)
        parsed_output[name] = value_tensor.squeeze(0)
    return parsed_output


def post_prepare_manual_inputs(
    prepared_inputs: Dict[str, Any],
    output_name: str,
    input_objects: "al_input_objects_as_dict",
) -> Dict[str, torch.Tensor]:
    prepared_inputs_post = {}

    for input_name, prepared_input in prepared_inputs.items():
        input_object = input_objects[input_name]
        input_info = input_object.input_config.input_info

        input_type = input_info.input_type
        if input_type != "sequence":
            prepared_inputs_post[input_name] = prepared_input
            continue

        assert isinstance(input_object, ComputedSequenceInputInfo)
        assert input_object.tokenizer is not None

        specials = get_special_tokens(
            tokenizer=input_object.tokenizer, vocab=input_object.vocab
        )
        pad_idx = specials.pad_idx

        if input_name == output_name:
            if input_info.input_type == "sequence":
                prepared_input_no_pad = [i for i in prepared_input if i != pad_idx]
                prepared_inputs_post[output_name] = torch.tensor(
                    prepared_input_no_pad,
                    dtype=torch.long,
                )
            else:
                raise NotImplementedError()

        else:
            prepared_inputs_post[input_name] = prepared_input

    return prepared_inputs_post


def _recursive_unsqueeze(
    data: dict,
    dim: int,
) -> Any:
    if isinstance(data, dict):
        return {k: _recursive_unsqueeze(v, dim=dim) for k, v in data.items()}
    elif isinstance(data, torch.Tensor):
        return data.unsqueeze(dim=dim)
    else:
        raise NotImplementedError()


def prepare_manual_sample_data(
    sample_inputs: Dict[str, Any], input_objects: "al_input_objects_as_dict"
) -> dict[str, torch.Tensor | dict[str, torch.Tensor]]:
    prepared_inputs: dict[str, torch.Tensor | dict[str, torch.Tensor]] = {}
    for name, data in sample_inputs.items():
        input_object = input_objects[name]
        input_info = input_object.input_config.input_info
        input_type_info = input_object.input_config.input_type_info

        match input_object:
            case ComputedOmicsInputInfo():
                assert isinstance(input_type_info, OmicsInputDataConfig)
                array_raw = omics_load_wrapper(
                    data_pointer=data,
                    input_source=input_info.input_source,
                    subset_indices=input_object.subset_indices,
                    deeplake_inner_key=input_info.input_inner_key,
                )

                array_prepared = prepare_one_hot_omics_data(
                    genotype_array=array_raw,
                    na_augment_alpha=input_type_info.na_augment_alpha,
                    na_augment_beta=input_type_info.na_augment_beta,
                    shuffle_augment_alpha=input_type_info.shuffle_augment_alpha,
                    shuffle_augment_beta=input_type_info.shuffle_augment_beta,
                    test_mode=True,
                )
                prepared_inputs[name] = array_prepared

            case ComputedSequenceInputInfo():
                assert isinstance(input_type_info, SequenceInputDataConfig)

                sequence_split = streamline_sequence_manual_data(
                    data=data,
                    split_on=input_type_info.split_on,
                )

                sequence_tokenized = input_object.encode_func(sequence_split)
                sequence_array = np.array(sequence_tokenized)

                prepared_sequence = prepare_sequence_data(
                    sequence_input_object=input_object,
                    cur_file_content_tokenized=sequence_array,
                    test_mode=True,
                )

                prepared_inputs[name] = prepared_sequence

            case ComputedBytesInputInfo():
                assert isinstance(input_type_info, ByteInputDataConfig)
                bytes_data = bytes_load_wrapper(
                    input_source=input_info.input_source,
                    data_pointer=data,
                    dtype=input_type_info.byte_encoding,
                    deeplake_inner_key=input_info.input_inner_key,
                )
                prepared_bytes_input = prepare_bytes_data(
                    bytes_input_object=input_object,
                    bytes_data=bytes_data,
                    test_mode=True,
                )

                prepared_inputs[name] = prepared_bytes_input

            case ComputedImageInputInfo():
                assert isinstance(input_type_info, ImageInputDataConfig)
                image_data = image_load_wrapper(
                    data_pointer=data,
                    image_mode=input_type_info.mode,
                    input_source=input_info.input_source,
                    deeplake_inner_key=input_info.input_inner_key,
                )
                prepared_image_data = prepare_image_data(
                    image_input_object=input_object,
                    image_data=image_data,
                    test_mode=True,
                )
                prepared_inputs[name] = prepared_image_data

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                assert isinstance(input_type_info, TabularInputDataConfig)
                transformers = input_object.labels.label_transformers
                tabular_data = _streamline_tabular_data_for_transformers(
                    tabular_input=data, transformers=transformers
                )
                prepared_inputs[name] = tabular_data

            case ComputedArrayInputInfo():
                assert isinstance(input_type_info, ArrayInputDataConfig)
                array_data = array_load_wrapper(
                    input_source=input_info.input_source,
                    data_pointer=data,
                    deeplake_inner_key=input_info.input_inner_key,
                )
                prepared_array_data = prepare_array_data(
                    array_data=array_data,
                    normalization_stats=input_object.normalization_stats,
                )
                prepared_inputs[name] = prepared_array_data

            case _:
                raise ValueError(
                    f"Unknown input type '{input_object.__class__.__name__}'"
                )

    return prepared_inputs


def serialize_raw_inputs(
    raw_inputs: dict[str, str | np.ndarray | dict[str, np.ndarray] | Image.Image],
    input_objects: "al_input_objects_as_dict",
    output_path: Path,
) -> None:
    warned_types: set[str] = set()
    for input_name, cur_input in raw_inputs.items():
        cur_input_object = input_objects[input_name]
        cur_input_type = cur_input_object.input_config.input_info.input_type
        cur_output_path = output_path / f"{input_name}"
        ensure_path_exists(path=cur_output_path)

        match cur_input_type:
            case "omics" | "array":
                assert isinstance(cur_input, np.ndarray)
                cur_output_path = cur_output_path.with_suffix(".npy")
                np.save(str(cur_output_path), cur_input)
            case "tabular":
                assert isinstance(cur_input, dict)
                cur_output_path = cur_output_path.with_suffix(".json")
                cur_output_path.write_text(data=json.dumps(cur_input))
            case "sequence":
                assert isinstance(cur_input, str)
                cur_output_path = cur_output_path.with_suffix(".txt")
                cur_output_path.write_text(data=cur_input)
            case "image":
                assert isinstance(cur_input, Image.Image)
                cur_output_path = cur_output_path.with_suffix(".png")
                cur_input.save(cur_output_path)
            case "bytes":
                assert isinstance(cur_input, np.ndarray)
                cur_output_path = cur_output_path.with_suffix(".bin")
                cur_input.tofile(cur_output_path)
            case _:
                if cur_input_type not in warned_types:
                    warned_types.add(cur_input_type)
                    logger.warning(
                        "Not serializing input of type %s. Not yet implemented.",
                        cur_input_type,
                    )


def get_dataset_loader_single_sample_generator(
    dataset: Dataset,
    infinite: bool = True,
) -> Iterator[al_getitem_return]:
    loader: Iterator[al_getitem_return]
    if infinite:
        loader = cycle(DataLoader(dataset=dataset, batch_size=1, shuffle=True))
    else:
        loader = iter(DataLoader(dataset=dataset, batch_size=1, shuffle=True))

    for input_to_model, _, cur_ids in loader:
        inputs_squeezed = _recursive_batch_dimension_squeeze(inputs=input_to_model)
        yield inputs_squeezed, {}, cur_ids


def _recursive_batch_dimension_squeeze(
    inputs: dict[str, torch.Tensor]
) -> dict[str, torch.Tensor]:
    """
    We need this function as when returning from a dataloader one sample at a time,
    the single samples include a batch dimension. When concatenating these samples
    we will end up with an extra dimension causing downstream issues.
    """

    inputs_squeezed = {}
    for name, data in inputs.items():
        if isinstance(data, dict):
            inputs_squeezed[name] = _recursive_batch_dimension_squeeze(data)
        else:
            assert data.shape[0] == 1
            inputs_squeezed[name] = data.squeeze(dim=0)

    return inputs_squeezed


def extract_input_types(input_objects: "al_input_objects_as_dict") -> dict[str, str]:
    input_types: dict[str, str] = {}
    for input_name, input_object in input_objects.items():
        input_types[input_name] = input_object.input_config.input_info.input_type
    return input_types


def prepare_base_input(
    prepared_inputs: dict[str, torch.Tensor],
) -> dict[str, Any]:
    prepared_sample_inputs_copy = deepcopy(prepared_inputs)
    return prepared_sample_inputs_copy


def get_batch_generator(
    iterator: Iterator[Tuple[int, Tuple[Any, Any]]],
    batch_size: int,
) -> Iterator[list[Tuple[int, Tuple[Any, Any]]]]:
    batch = []
    for item in iterator:
        batch.append(item)
        if len(batch) == batch_size:
            yield batch
            batch = []
    if batch:
        yield batch



from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Sequence,
    Tuple,
    Union,
    overload,
)

import torch
from torch.utils.data import DistributedSampler, WeightedRandomSampler

from eir.data_load.data_loading_funcs import get_weighted_random_sampler
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schema_modules.output_schemas_survival import SurvivalOutputTypeConfig
from eir.train_utils.distributed import in_distributed_env
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.datasets import al_local_datasets, al_sample_label_dict_target
    from eir.setup.output_setup import al_output_objects_as_dict

logger = get_logger(name=__name__)


def get_output_info_generator(
    outputs_as_dict: "al_output_objects_as_dict",
) -> Generator[Tuple[str, str, str], None, None]:
    for output_name, output_object in outputs_as_dict.items():
        match output_object:
            case ComputedTabularOutputInfo():
                target_columns = output_object.target_columns
                for column_type, list_of_cols_of_this_type in target_columns.items():
                    for cur_column in list_of_cols_of_this_type:
                        yield output_name, column_type, cur_column
            case (
                ComputedSequenceOutputInfo()
                | ComputedArrayOutputInfo()
                | ComputedImageOutputInfo()
            ):
                yield output_name, "general", output_name
            case ComputedSurvivalOutputInfo():
                output_type_info = output_object.output_config.output_type_info
                assert isinstance(output_type_info, SurvivalOutputTypeConfig)
                event_name = output_type_info.event_column
                yield output_name, "survival", event_name
            case _:
                raise TypeError(f"Unknown output object: {output_object}")


@dataclass(frozen=True)
class Batch:
    inputs: Dict[str, torch.Tensor]
    target_labels: Dict[str, Dict[str, torch.Tensor]]
    ids: List[str]


@overload
def get_train_sampler(
    columns_to_sample: None, train_dataset: "al_local_datasets"
) -> None: ...


@overload
def get_train_sampler(
    columns_to_sample: Sequence[str],
    train_dataset: "al_local_datasets",
) -> Union[WeightedRandomSampler, DistributedSampler]: ...


def get_train_sampler(columns_to_sample, train_dataset):
    in_distributed_run = in_distributed_env()

    if columns_to_sample is None:
        if in_distributed_run:
            return DistributedSampler(dataset=train_dataset)
        else:
            return None

    if in_distributed_run:
        raise NotImplementedError(
            "Weighted sampling not yet implemented for distributed training."
        )

    logger.debug("Weighted sampling enabled, setting up.")

    loaded_target_columns = _gather_all_loaded_columns(outputs=train_dataset.outputs)

    is_sample_column_loaded = False
    is_sample_all_cols = False

    if columns_to_sample == ["all"]:
        is_sample_all_cols = True
    else:
        parsed_columns_to_sample = set(i.split("__", 1)[1] for i in columns_to_sample)
        is_sample_column_loaded = parsed_columns_to_sample.issubset(
            set(loaded_target_columns)
        )

    if not is_sample_column_loaded and not is_sample_all_cols:
        raise ValueError(
            "Weighted sampling from non-loaded columns not supported yet "
            f"(could not find {columns_to_sample})."
        )

    if is_sample_all_cols:
        columns_to_sample = []
        for output_name, output_object in train_dataset.outputs.items():
            if output_object.output_config.output_info.output_type != "tabular":
                continue

            cat_columns_with_output_prefix = [
                output_name + "__" + i for i in output_object.target_columns["cat"]
            ]
            columns_to_sample += cat_columns_with_output_prefix

    train_sampler = get_weighted_random_sampler(
        target_storage=train_dataset.target_labels_storage,
        columns_to_sample=columns_to_sample,
    )
    return train_sampler


def _gather_all_loaded_columns(outputs: "al_output_objects_as_dict") -> Sequence[str]:
    loaded_cat_columns: list[str] = []
    loaded_con_columns: list[str] = []

    for output_name, output_object in outputs.items():
        match output_object:
            case ComputedTabularOutputInfo():
                cur_target_columns = output_object.target_columns
                loaded_cat_columns += cur_target_columns.get("cat", [])
                loaded_con_columns += cur_target_columns.get("con", [])
            case ComputedSequenceOutputInfo():
                pass
            case _:
                raise TypeError(f"Unknown output object: {output_object}")

    loaded_target_columns = loaded_cat_columns + loaded_con_columns

    return loaded_target_columns


@dataclass
class Sample:
    sample_id: str
    inputs: dict[str, Any]
    target_labels: "al_sample_label_dict_target"



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy as np

from eir.setup.input_setup_modules.common import DataDimensions
from eir.setup.input_setup_modules.setup_array import (
    ArrayNormalizationStats,
    get_array_normalization_values,
    get_data_dimension_from_data_source,
    get_dtype_from_data_source,
)
from eir.setup.schemas import ArrayOutputTypeConfig, OutputConfig
from eir.train_utils.step_modules.diffusion import (
    DiffusionConfig,
    initialize_diffusion_config,
)
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass
class ComputedArrayOutputInfo:
    output_config: OutputConfig
    data_dimensions: DataDimensions
    dtype: np.dtype
    normalization_stats: Optional[ArrayNormalizationStats] = None
    diffusion_config: Optional[DiffusionConfig] = None


def set_up_array_output(
    output_config: OutputConfig,
    normalization_stats: Optional[ArrayNormalizationStats] = None,
    data_dimensions: Optional[DataDimensions] = None,
    diffusion_config: Optional[DiffusionConfig] = None,
    dtype: Optional[np.dtype] = None,
    *args,
    **kwargs,
) -> ComputedArrayOutputInfo:

    if data_dimensions is None:
        data_dimensions = get_data_dimension_from_data_source(
            data_source=Path(output_config.output_info.output_source),
            deeplake_inner_key=output_config.output_info.output_inner_key,
        )

    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, ArrayOutputTypeConfig)

    if normalization_stats is None:
        if output_type_info.normalization is not None:
            normalization_stats = get_array_normalization_values(
                source=output_config.output_info.output_source,
                inner_key=output_config.output_info.output_inner_key,
                normalization=output_type_info.normalization,
                data_dimensions=data_dimensions,
                max_samples=output_type_info.adaptive_normalization_max_samples,
            )

    if dtype is None:
        dtype = get_dtype_from_data_source(
            data_source=Path(output_config.output_info.output_source),
            deeplake_inner_key=output_config.output_info.output_inner_key,
        )

    if diffusion_config is None:
        if output_type_info.loss == "diffusion":
            time_steps = output_type_info.diffusion_time_steps
            if time_steps is None:
                raise ValueError(
                    "Diffusion loss requires specifying the number of time steps."
                    "Please set `diffusion_time_steps` in the output config."
                )
            diffusion_config = initialize_diffusion_config(time_steps=time_steps)

    array_output_object = ComputedArrayOutputInfo(
        output_config=output_config,
        data_dimensions=data_dimensions,
        normalization_stats=normalization_stats,
        dtype=dtype,
        diffusion_config=diffusion_config,
    )

    return array_output_object



from contextlib import nullcontext
from dataclasses import dataclass
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import torch
from ignite.engine import Engine
from torch import autocast, nn
from torch.cuda.amp import GradScaler
from torch.nn.utils import clip_grad_norm_
from torch.optim.optimizer import Optimizer

from eir.data_load.data_augmentation import get_mix_data_hook, hook_mix_loss
from eir.data_load.data_utils import Batch
from eir.models import model_training_utils
from eir.models.input.tabular.tabular import get_tabular_inputs
from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo
from eir.setup import schemas
from eir.setup.config import Configs, get_all_tabular_targets
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.setup_array import ComputedArrayInputInfo
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.output_setup import al_output_objects_as_dict
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.train_utils.metrics import (
    add_loss_to_metrics,
    add_multi_task_average_metrics,
    aggregate_losses,
    calculate_batch_metrics,
    filter_missing_outputs_and_labels,
    get_uncertainty_loss_hook,
    hook_add_l1_loss,
)
from eir.train_utils.optim import AttrDelegatedSWAWrapper, get_optimizer_backward_kwargs
from eir.train_utils.step_modules.autoregressive import (
    prepare_sequence_input_for_sequence_output,
)
from eir.train_utils.step_modules.diffusion import prepare_diffusion_batch
from eir.train_utils.train_handlers import HandlerConfig
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.model_setup_modules.meta_setup import al_meta_model
    from eir.train import Experiment

al_training_labels_target = Dict[str, Dict[str, torch.Tensor]]
al_input_batch = Dict[str, torch.Tensor]
al_ids = List[str]
al_dataloader_getitem_batch = Tuple[
    al_input_batch,
    al_training_labels_target,
    al_ids,
]

logger = get_logger(name=__name__)


def get_default_hooks(configs: Configs) -> "Hooks":
    step_func_hooks, extra_state = _get_default_step_function_hooks(configs=configs)
    hooks_object = Hooks(step_func_hooks=step_func_hooks, extra_state=extra_state)

    return hooks_object


@dataclass
class Hooks:
    al_handler_attachers = Iterable[Callable[[Engine, HandlerConfig], Engine]]

    step_func_hooks: "StepFunctionHookStages"
    extra_state: Optional[dict[str, Any]] = None


def _get_default_step_function_hooks(
    configs: Configs,
) -> tuple["StepFunctionHookStages", dict[str, Any]]:
    """
    TODO: Add validation, inspect that outputs have correct names.
    TODO: Refactor, split into smaller functions e.g. for L1, mixing and uncertainty.
    """

    init_kwargs, extra_state = _get_default_step_function_hooks_init_kwargs(
        configs=configs
    )

    step_func_hooks = StepFunctionHookStages(**init_kwargs)

    return step_func_hooks, extra_state


def _get_default_step_function_hooks_init_kwargs(
    configs: Configs,
) -> tuple[dict[str, list[Callable]], dict[str, Any]]:
    init_kwargs: dict[str, list[Callable]] = {
        "base_prepare_batch": [hook_default_prepare_batch],
        "post_prepare_batch": [],
        "model_forward": [hook_default_model_forward],
        "loss": [hook_default_per_target_loss],
        "optimizer_backward": [hook_default_optimizer_backward],
        "metrics": [hook_default_compute_metrics],
    }

    extra_state: dict[str, Any] = {}

    if configs.gc.tc.mixing_alpha:
        logger.debug(
            "Setting up hooks for mixing with with α=%.2g.",
            configs.gc.tc.mixing_alpha,
        )
        mix_hook = get_mix_data_hook(input_configs=configs.input_configs)

        init_kwargs["post_prepare_batch"].append(mix_hook)
        init_kwargs["loss"][0] = hook_mix_loss

    if _should_add_uncertainty_loss_hook(output_configs=configs.output_configs):
        uncertainty_hook, uncertainty_modules = get_uncertainty_loss_hook(
            output_configs=configs.output_configs,
            device=configs.gc.be.device,
        )
        init_kwargs["loss"].append(uncertainty_hook)
        extra_state["uncertainty_modules"] = uncertainty_modules

    init_kwargs["loss"].append(hook_default_aggregate_losses)

    init_kwargs = add_l1_loss_hook_if_applicable(
        step_function_hooks_init_kwargs=init_kwargs, configs=configs
    )

    grad_acc_steps = configs.gc.opt.gradient_accumulation_steps
    if grad_acc_steps and grad_acc_steps > 1:
        logger.debug(
            "Adding gradient accumulation hook with steps=%d.",
            configs.gc.opt.gradient_accumulation_steps,
        )
    init_kwargs["loss"].append(get_hook_iteration_counter())

    do_amp = configs.gc.m.amp
    if do_amp:
        logger.debug("Setting up AMP training.")
        model_forward_with_amp_objects = [
            get_hook_amp_objects(device=configs.gc.be.device)
        ] + init_kwargs["model_forward"]
        init_kwargs["model_forward"] = model_forward_with_amp_objects

    return init_kwargs, extra_state


def _should_add_uncertainty_loss_hook(
    output_configs: Sequence[schemas.OutputConfig],
) -> bool:
    all_targets = get_all_tabular_targets(output_configs=output_configs)

    more_than_one_target = len(all_targets) > 1

    any_uncertainty_targets = any(
        hasattr(c.output_type_info, "uncertainty_weighted_mt_loss")
        and c.output_type_info.uncertainty_weighted_mt_loss
        for c in output_configs
    )
    return more_than_one_target and any_uncertainty_targets


def add_l1_loss_hook_if_applicable(
    step_function_hooks_init_kwargs: Dict[str, list[Callable]],
    configs: Configs,
) -> Dict[str, List[Callable]]:
    input_l1 = any(
        getattr(input_config.model_config.model_init_config, "l1", None)
        for input_config in configs.input_configs
    )
    fusion_l1 = getattr(configs.fusion_config.model_config, "l1", None)
    if input_l1 or fusion_l1:
        logger.debug("Adding L1 loss hook.")
        step_function_hooks_init_kwargs["loss"].append(hook_add_l1_loss)

    return step_function_hooks_init_kwargs


@dataclass
class StepFunctionHookStages:
    al_hook = Callable[..., dict]
    al_hooks = Iterable[al_hook]

    base_prepare_batch: al_hooks
    post_prepare_batch: al_hooks
    model_forward: al_hooks
    loss: al_hooks
    optimizer_backward: al_hooks
    metrics: al_hooks


def hook_default_prepare_batch(
    experiment: "Experiment",
    loader_batch: al_dataloader_getitem_batch,
    *args,
    **kwargs,
) -> Dict:
    batch = prepare_base_batch_default(
        loader_batch=loader_batch,
        input_objects=experiment.inputs,
        output_objects=experiment.outputs,
        model=experiment.model,
        device=experiment.configs.gc.be.device,
    )

    state_updates = {"batch": batch}

    return state_updates


def prepare_base_batch_default(
    loader_batch: al_dataloader_getitem_batch,
    input_objects: al_input_objects_as_dict,
    output_objects: al_output_objects_as_dict,
    model: nn.Module,
    device: str,
) -> Batch:
    inputs, target_labels, ids = loader_batch

    inputs_prepared, generated_targets = _prepare_inputs_for_model(
        batch_inputs=inputs,
        input_objects=input_objects,
        output_objects=output_objects,
        model=model,
        device=device,
    )

    if not generated_targets:
        generated_targets = {}
    if not target_labels:
        target_labels = {}
    else:
        target_labels = model_training_utils.prepare_all_targets(
            output_objects=output_objects,
            device=device,
            labels=target_labels,
        )

    target_labels = {**target_labels, **generated_targets}

    batch = Batch(
        inputs=inputs_prepared,
        target_labels=target_labels,
        ids=ids,
    )

    return batch


def _prepare_inputs_for_model(
    batch_inputs: Dict[str, Any],
    input_objects: al_input_objects_as_dict,
    output_objects: al_output_objects_as_dict,
    model: nn.Module,
    device: str,
) -> Tuple[al_input_batch, al_training_labels_target]:
    inputs_prepared = {}
    targets_prepared = {}

    for input_name, input_object in input_objects.items():
        match input_object:
            case ComputedOmicsInputInfo():
                cur_tensor = batch_inputs[input_name]
                cur_tensor = cur_tensor.to(dtype=torch.float32)
                cur_tensor = cur_tensor.to(device=device)

                inputs_prepared[input_name] = cur_tensor

            case ComputedArrayInputInfo() | ComputedImageInputInfo():
                cur_tensor = batch_inputs[input_name]
                cur_tensor = cur_tensor.to(dtype=torch.float32)
                cur_tensor = cur_tensor.to(device=device)

                if input_name in (i for i in output_objects.keys()):
                    matching_output = output_objects[input_name]
                    assert isinstance(
                        matching_output,
                        (ComputedArrayOutputInfo, ComputedImageOutputInfo),
                    )
                    output_config = matching_output.output_config
                    output_type_info = output_config.output_type_info
                    assert isinstance(
                        output_type_info,
                        (schemas.ArrayOutputTypeConfig, schemas.ImageOutputTypeConfig),
                    )
                    loss = output_type_info.loss

                    if loss == "diffusion":
                        assert matching_output.diffusion_config is not None
                        num_steps = output_type_info.diffusion_time_steps
                        assert num_steps is not None
                        cur_tensor, cur_target, t = prepare_diffusion_batch(
                            diffusion_config=matching_output.diffusion_config,
                            inputs=cur_tensor,
                            batch_size=cur_tensor.shape[0],
                            num_steps=num_steps,
                        )
                        cur_targets = {input_name: cur_target}
                        targets_prepared[input_name] = cur_targets
                        output_module = getattr(model.output_modules, input_name)
                        t_emb = output_module.feature_extractor.timestep_embeddings(t)
                        inputs_prepared[f"__extras_{input_name}"] = t_emb

                inputs_prepared[input_name] = cur_tensor

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                tabular_source_input = batch_inputs[input_name]
                for tabular_name, tensor in tabular_source_input.items():
                    if torch.is_floating_point(input=tensor):
                        tensor = tensor.to(dtype=torch.float32)

                    cur_tensor = tensor.to(device=device)
                    tabular_source_input[tabular_name] = cur_tensor

                tabular_input_type_info = input_object.input_config.input_type_info
                assert isinstance(
                    tabular_input_type_info, schemas.TabularInputDataConfig
                )
                cat_columns = tabular_input_type_info.input_cat_columns
                con_columns = tabular_input_type_info.input_con_columns
                tabular = get_tabular_inputs(
                    input_cat_columns=cat_columns,
                    input_con_columns=con_columns,
                    tabular_model=getattr(model.input_modules, input_name),
                    tabular_input=tabular_source_input,
                    device=device,
                )
                inputs_prepared[input_name] = tabular

            case ComputedSequenceInputInfo():
                cur_seq = batch_inputs[input_name]

                if input_name in (i for i in output_objects.keys()):
                    cur_seq, cur_targets = prepare_sequence_input_for_sequence_output(
                        input_object=input_object,
                        cur_seq=cur_seq,
                        input_name=input_name,
                        device=device,
                    )
                    targets_prepared[input_name] = cur_targets

                cur_embedding = _prepare_sequence_input_base(
                    cur_seq=cur_seq,
                    input_name=input_name,
                    model=model,
                    device=device,
                )
                inputs_prepared[input_name] = cur_embedding

            case ComputedBytesInputInfo():
                cur_seq = batch_inputs[input_name]

                cur_embedding = _prepare_sequence_input_base(
                    cur_seq=cur_seq,
                    input_name=input_name,
                    model=model,
                    device=device,
                )
                inputs_prepared[input_name] = cur_embedding

            case _:
                raise ValueError(f"Unrecognized input type {input_name}.")

    return inputs_prepared, targets_prepared


def _prepare_sequence_input_base(
    cur_seq: torch.Tensor,
    model: nn.Module,
    input_name: str,
    device: str,
) -> torch.Tensor:
    cur_seq = cur_seq.to(device=device)
    cur_module = getattr(model.input_modules, input_name)
    cur_module_embedding = cur_module.embedding
    cur_embedding = cur_module_embedding(input=cur_seq)

    return cur_embedding


def hook_default_model_forward(
    experiment: "Experiment", state: Dict, batch: "Batch", *args, **kwargs
) -> Dict:
    inputs = batch.inputs

    context_manager = get_maybe_amp_context_manager_from_state(state=state)
    with context_manager:
        train_outputs = experiment.model(inputs=inputs)

    state_updates = {"model_outputs": train_outputs}

    return state_updates


def get_amp_context_manager(device_type: str) -> autocast:
    return autocast(device_type=device_type)


def hook_default_optimizer_backward(
    experiment: "Experiment", state: Dict, *args, **kwargs
) -> Dict:
    gc = experiment.configs.global_config
    optimizer_backward_kwargs = get_optimizer_backward_kwargs(
        optimizer_name=gc.opt.optimizer
    )

    loss = maybe_scale_loss_with_grad_accumulation_steps(
        loss=state["loss"],
        grad_acc_steps=gc.opt.gradient_accumulation_steps,
    )

    amp_scaler = state.get("amp_scaler")
    loss = maybe_scale_loss_with_amp_scaler(
        do_amp=gc.m.amp,
        loss=loss,
        amp_scaler=amp_scaler,
        device=gc.be.device,
    )

    loss.backward(**optimizer_backward_kwargs)

    maybe_apply_gradient_noise_to_model(
        model=experiment.model,
        gradient_noise=gc.opt.gradient_noise,
    )
    maybe_apply_gradient_clipping_to_model(
        model=experiment.model,
        gradient_clipping=gc.opt.gradient_clipping,
    )

    step_func = get_optimizer_step_func(
        do_amp=gc.m.amp,
        optimizer=experiment.optimizer,
        amp_scaler=amp_scaler,
        device=gc.be.device,
    )

    if should_perform_optimizer_step(
        iteration=state["iteration"],
        grad_acc_steps=gc.opt.gradient_accumulation_steps,
    ):
        step_func()

    if gc.m.amp and gc.be.device != "cpu":
        assert amp_scaler is not None
        amp_scaler.update()

    maybe_update_model_parameters_with_swa(
        n_iter_before_swa=gc.m.n_iter_before_swa,
        model=experiment.model,
        iteration=state["iteration"],
        sample_interval=gc.ec.sample_interval,
    )

    return {}


def maybe_scale_loss_with_amp_scaler(
    do_amp: bool, loss: torch.Tensor, amp_scaler: Optional["GradScaler"], device: str
) -> torch.Tensor:
    if do_amp and device != "cpu":
        assert amp_scaler is not None
        return amp_scaler.scale(loss)
    else:
        return loss


def maybe_scale_loss_with_grad_accumulation_steps(
    loss: torch.Tensor, grad_acc_steps: Optional[int]
) -> torch.Tensor:
    if grad_acc_steps and grad_acc_steps > 1:
        return loss / grad_acc_steps
    else:
        return loss


def maybe_apply_gradient_noise_to_model(
    model: "nn.Module", gradient_noise: float
) -> None:
    if gradient_noise:
        for name, weight in model.named_parameters():
            if weight.grad is None:
                continue

            weight.grad = weight.grad + torch.randn_like(weight.grad) * gradient_noise


def maybe_apply_gradient_clipping_to_model(
    model: "nn.Module", gradient_clipping: float
) -> None:
    if gradient_clipping:
        clip_grad_norm_(
            parameters=model.parameters(),
            max_norm=gradient_clipping,
        )


def get_optimizer_step_func(
    do_amp: bool,
    optimizer: Optimizer,
    amp_scaler: Optional["GradScaler"],
    device: str,
) -> Callable:
    if do_amp and device != "cpu":
        assert amp_scaler is not None
        return partial(amp_scaler.step, optimizer=optimizer)
    else:
        return optimizer.step


def should_perform_optimizer_step(
    iteration: int, grad_acc_steps: Optional[int]
) -> bool:
    if grad_acc_steps and grad_acc_steps > 1:
        return iteration % grad_acc_steps == 0
    else:
        return True


def maybe_update_model_parameters_with_swa(
    n_iter_before_swa: Optional[int],
    model: "al_meta_model",
    iteration: int,
    sample_interval: int,
) -> None:
    should_not_be_called_ever = n_iter_before_swa is None
    if should_not_be_called_ever:
        return

    assert isinstance(model, AttrDelegatedSWAWrapper)
    assert n_iter_before_swa is not None
    if iteration >= n_iter_before_swa and iteration % sample_interval == 0:
        model.update_parameters(model.module)


def hook_default_compute_metrics(
    experiment: "Experiment", batch: "Batch", state: Dict, *args, **kwargs
) -> dict[str, Any]:

    filtered_outputs = state["filtered_outputs"]

    train_batch_metrics = calculate_batch_metrics(
        outputs_as_dict=experiment.outputs,
        outputs=filtered_outputs.model_outputs,
        labels=filtered_outputs.target_labels,
        mode="train",
        metric_record_dict=experiment.metrics,
    )

    train_batch_metrics_w_loss = add_loss_to_metrics(
        outputs_as_dict=experiment.outputs,
        losses=state["per_target_train_losses"],
        metric_dict=train_batch_metrics,
    )

    train_batch_metrics_with_averages = add_multi_task_average_metrics(
        batch_metrics_dict=train_batch_metrics_w_loss,
        outputs_as_dict=experiment.outputs,
        loss=state["loss"].item(),
        performance_average_functions=None,
    )

    state_updates = {"metrics": train_batch_metrics_with_averages}

    return state_updates


def hook_default_per_target_loss(
    experiment: "Experiment",
    batch: "Batch",
    state: Dict,
    *args,
    **kwargs,
) -> dict[str, Any]:
    """
    Note that we do not filter within modality missing sample data here
    after `filter_missing_outputs_and_labels` as that is handled in
    the respective (tabular and survival output) loss functions.
    """
    context_manager = get_maybe_amp_context_manager_from_state(state=state)
    with context_manager:
        model_outputs = state["model_outputs"]
        target_labels = batch.target_labels
        ids = batch.ids

        filtered_outputs = filter_missing_outputs_and_labels(
            batch_ids=ids,
            model_outputs=model_outputs,
            target_labels=target_labels,
            missing_ids_info=experiment.valid_dataset.missing_ids_per_output,
            with_labels=True,
        )

        per_target_train_losses = experiment.loss_function(
            inputs=filtered_outputs.model_outputs,
            targets=filtered_outputs.target_labels,
        )

        state_updates = {
            "per_target_train_losses": per_target_train_losses,
            "filtered_outputs": filtered_outputs,
        }

    return state_updates


def hook_default_aggregate_losses(state: Dict, *args, **kwargs) -> Dict:
    context_manager = get_maybe_amp_context_manager_from_state(state=state)
    with context_manager:
        train_loss_avg = aggregate_losses(losses_dict=state["per_target_train_losses"])
        state_updates = {"loss": train_loss_avg}

    return state_updates


def get_maybe_amp_context_manager_from_state(
    state: Dict,
) -> Union[nullcontext, autocast]:
    context_manager = state.get("amp_context_manager", nullcontext())
    return context_manager


def get_hook_iteration_counter() -> Callable:
    iteration_count = 0

    def _counter_iterator(do_increment: bool = True, *args, **kwargs) -> Dict[str, int]:
        nonlocal iteration_count
        if do_increment:
            iteration_count += 1

        state_updates = {"iteration": iteration_count}
        return state_updates

    return _counter_iterator


def get_hook_amp_objects(device: str) -> Callable[..., Dict[str, Any]]:
    device_type = "cpu" if device == "cpu" else "cuda"

    if device == "cpu":
        logger.warning("Using AMP is on a CPU, speedups will most likely be minimal.")

    scaler = None
    if device != "cpu":
        scaler = GradScaler()

    amp_context_manager = get_amp_context_manager(device_type=device_type)

    def _get_objects(*args, **kwargs) -> dict[str, GradScaler | autocast]:
        state_updates: dict[str, autocast | GradScaler] = {
            "amp_context_manager": amp_context_manager,
        }
        if device != "cpu":
            assert scaler is not None
            state_updates["amp_scaler"] = scaler

        return state_updates

    return _get_objects



from typing import Any, Sequence

from fastapi import FastAPI

from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_api_models import ResponseModel, create_input_model
from eir.serve_modules.serve_experiment_io import ServeExperiment
from eir.serve_modules.serve_input_setup import general_pre_process
from eir.serve_modules.serve_post_process import general_post_process
from eir.serve_modules.serve_prediction import run_serve_prediction
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.setup_array import ComputedArrayInputInfo
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.output_setup import al_output_objects_as_dict
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schemas import InputConfig
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)


async def process_request(data: Sequence, serve_experiment: ServeExperiment) -> Any:
    batch = general_pre_process(
        data=data,
        serve_experiment=serve_experiment,
    )

    predictions = run_serve_prediction(
        serve_experiment=serve_experiment,
        batch=batch,
    )

    responses = []
    for prediction in predictions:
        response = general_post_process(
            outputs=prediction,
            input_objects=serve_experiment.inputs,
            output_objects=serve_experiment.outputs,
        )
        responses.append(response)

    return responses


def create_predict_endpoint(
    app: FastAPI,
    configs: Sequence[InputConfig],
    serve_experiment: ServeExperiment,
) -> None:
    input_model = create_input_model(configs=configs)

    @app.post("/predict", response_model=ResponseModel)
    async def predict(requests: list[input_model]) -> ResponseModel:  # type: ignore
        logger.info(f"Received {len(requests)} samples in request.")
        data = [request.model_dump() for request in requests]  # type: ignore
        response_data = await process_request(
            data=data,
            serve_experiment=serve_experiment,
        )

        return ResponseModel(result=response_data)


def create_info_endpoint(app: FastAPI, serve_experiment: ServeExperiment) -> None:
    model_info: dict[str, Any] = get_model_info(
        input_objects=serve_experiment.inputs,
        output_objects=serve_experiment.outputs,
    )

    @app.get("/info")
    def info() -> dict[str, Any]:
        return model_info


def get_model_info(
    input_objects: al_input_objects_as_dict,
    output_objects: al_output_objects_as_dict,
) -> dict[str, Any]:
    model_info: dict[str, Any] = {
        "inputs": {},
        "outputs": {},
    }

    for name, input_object in input_objects.items():
        match input_object:
            case ComputedOmicsInputInfo():
                shape = input_object.data_dimensions.full_shape()[1:]
                model_info["inputs"][name] = {
                    "type": "omics",
                    "shape": shape,
                }

            case ComputedSequenceInputInfo() | ComputedBytesInputInfo():
                pass

            case ComputedImageInputInfo():
                shape = input_object.data_dimensions.full_shape()
                model_info["inputs"][name] = {
                    "type": "image",
                    "shape": shape,
                }

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                pass

            case ComputedArrayInputInfo():
                shape = input_object.data_dimensions.full_shape()
                model_info["inputs"][name] = {
                    "type": "array",
                    "shape": shape,
                    "dtype": input_object.dtype.str,
                }

            case _:
                input_type = input_object.input_config.input_info.input_type
                raise ValueError(f"Unknown input type: {input_type}")

    for name, output_object in output_objects.items():
        match output_object:
            case ComputedTabularOutputInfo():
                pass

            case ComputedSequenceOutputInfo():
                pass

            case ComputedArrayOutputInfo():
                shape = output_object.data_dimensions.full_shape()
                model_info["outputs"][name] = {
                    "type": "array",
                    "shape": shape,
                    "dtype": output_object.dtype.str,
                }

            case ComputedImageOutputInfo():
                shape = output_object.data_dimensions.full_shape()
                model_info["outputs"][name] = {
                    "type": "image",
                    "shape": shape,
                }

            case ComputedSurvivalOutputInfo():
                pass

            case _:
                output_type = output_object.output_config.output_info.output_type
                raise ValueError(f"Unknown output type: {output_type}")

    return model_info
